"""
This module runs experiment of IDW method

@author: Jakub Walczak, PhD
"""

import shutil
from collections import defaultdict
from pathlib import Path

import pandas as pd
import xarray as xr
from rich.console import Console

import climatrix as cm

console = Console()

INF_LOSS = -1e4

# Setting up the experiment parameters
NAN_POLICY = "resample"
console.print("[bold green]Using NaN policy: [/bold green]", NAN_POLICY)

SEED = 1
console.print("[bold green]Using seed: [/bold green]", SEED)

DSET_PATH = Path(__file__).parent.parent.parent.joinpath("data")
console.print("[bold green]Using dataset path: [/bold green]", DSET_PATH)

OPTIM_N_ITERS: int = 100
console.print(
    "[bold green]Using iterations for optimization[/bold green]", OPTIM_N_ITERS
)

RESULT_DIR: Path = Path(__file__).parent.parent.parent / "results" / "idw"
PLOT_DIR: Path = RESULT_DIR / "plots"
PLOT_DIR.mkdir(parents=True, exist_ok=True)
console.print("[bold green]Plots will be saved to: [/bold green]", PLOT_DIR)

METRICS_PATH: Path = RESULT_DIR / "metrics.csv"
console.print(
    "[bold green]Metrics will be saved to: [/bold green]", METRICS_PATH
)

HYPERPARAMETERS_SUMMARY_PATH: Path = RESULT_DIR / "hparams_summary.csv"
console.print(
    "[bold green]Hyperparameters summary will be saved to: [/bold green]",
    HYPERPARAMETERS_SUMMARY_PATH,
)

BOUNDS = {
    "k": (1, 50),
    "power": (1e-7, 5.0),
    "k_min": (1, 40),
}
console.print("[bold green]Hyperparameter bounds: [/bold green]", BOUNDS)

EUROPE_BOUNDS = {"north": 71, "south": 36, "west": -24, "east": 35}
EUROPE_DOMAIN = cm.Domain.from_lat_lon(
    lat=slice(EUROPE_BOUNDS["south"], EUROPE_BOUNDS["north"], 0.1),
    lon=slice(EUROPE_BOUNDS["west"], EUROPE_BOUNDS["east"], 0.1),
    kind="dense",
)


def clear_result_dir():
    console.print(
        "[bold red]Clearing result directory for this experiment...[/bold red]"
    )
    shutil.rmtree(RESULT_DIR, ignore_errors=True)


def create_result_dir():
    PLOT_DIR.mkdir(parents=True, exist_ok=True)


def get_all_dataset_idx() -> list[str]:
    return sorted(
        list({path.stem.split("_")[-1] for path in DSET_PATH.glob("*.nc")})
    )


def run_experiment():
    dset_idx = get_all_dataset_idx()
    with console.status("[magenta]Preparing experiment...") as status:
        all_metrics = {}
        hyperparams = defaultdict(list)
        for i, d in enumerate(dset_idx):
            cm.seed_all(SEED)
            status.update(
                f"[magenta]Processing date: {d} ({i + 1}/{len(dset_idx)})...",
                spinner="bouncingBall",
            )
            train_dset = xr.open_dataset(
                DSET_PATH / f"ecad_obs_europe_train_{d}.nc"
            ).cm
            val_dset = xr.open_dataset(
                DSET_PATH / f"ecad_obs_europe_val_{d}.nc"
            ).cm
            test_dset = xr.open_dataset(
                DSET_PATH / f"ecad_obs_europe_test_{d}.nc"
            ).cm
            status.update(
                f"[magenta]Optimizing hyper-parameters for date: {d}"
                f" ({i + 1}/{len(dset_idx)})...",
                spinner="bouncingBall",
            )
            finder = cm.optim.HParamFinder(
                "idw",
                train_dset,
                val_dset,
                metric="mae",
                explore=0.3,
                n_iters=OPTIM_N_ITERS,
                bounds=BOUNDS,
                random_seed=SEED,
            )
            result = finder.optimize()
            console.print("[bold yellow]Optimized parameters:[/bold yellow]")
            console.print(
                "[yellow]Power:[/yellow]", result["best_params"]["power"]
            )
            console.print("[yellow]k:[/yellow]", result["best_params"]["k"])
            console.print(
                "[yellow]k_min:[/yellow]", result["best_params"]["k_min"]
            )
            status.update(
                "[magenta]Reconstructing with optimised parameters...",
                spinner="bouncingBall",
            )
            status.update(
                "[magenta]Concatenating train and validation datasets...",
                spinner="bouncingBall",
            )
            train_val_dset = xr.concat(
                [train_dset.da, val_dset.da], dim="point"
            ).cm
            reconstructed_dset = train_val_dset.reconstruct(
                test_dset.domain,
                method="idw",
                k=result["best_params"]["k"],
                power=result["best_params"]["power"],
                k_min=result["best_params"]["k_min"],
            )
            status.update(
                "[magenta]Saving reconstructed dset to "
                f"{PLOT_DIR}/{d}_reconstructed.png...",
                spinner="bouncingBall",
            )
            reconstructed_dset.plot(show=False).get_figure().savefig(
                PLOT_DIR / f"{d}_reconstructed.png"
            )

            status.update(
                "[magenta]Reconstructing to dense Europe domain...",
                spinner="bouncingBall",
            )
            reconstructed_dense = train_val_dset.reconstruct(
                EUROPE_DOMAIN,
                method="idw",
                k=result["best_params"]["k"],
                power=result["best_params"]["power"],
                k_min=result["best_params"]["k_min"],
            )
            status.update(
                "[magenta]Saving reconstructed dense dset to "
                f"{PLOT_DIR}/{d}_reconstructed_dense.png...",
                spinner="bouncingBall",
            )
            reconstructed_dense.plot(show=False).get_figure().savefig(
                PLOT_DIR / f"{d}_reconstructed_dense.png"
            )
            status.update(
                "[magenta]Saving test dset to " f"{PLOT_DIR}/{d}_test.png...",
                spinner="bouncingBall",
            )
            test_dset.plot(show=False).get_figure().savefig(
                PLOT_DIR / f"{d}_test.png"
            )
            status.update("[magenta]Evaluating...", spinner="bouncingBall")
            cmp = cm.Comparison(reconstructed_dset, test_dset)
            cmp.diff.plot(show=False).get_figure().savefig(
                PLOT_DIR / f"{d}_diffs.png"
            )
            cmp.plot_signed_diff_hist().get_figure().savefig(
                PLOT_DIR / f"{d}_hist.png"
            )
            metrics = cmp.compute_report()
            metrics["dataset_id"] = d
            all_metrics.append(metrics)

            hyperparams["dataset_id"].append(d)
            hyperparams["k"].append(result["best_params"]["k"])
            hyperparams["power"].append(result["best_params"]["power"])
            hyperparams["k_min"].append(result["best_params"]["k_min"])
            hyperparams["opt_loss"].append(result["best_score"])
        status.update(
            "[magenta]Saving quality metrics...", spinner="bouncingBall"
        )
        pd.DataFrame(all_metrics).to_csv(METRICS_PATH)
        status.update(
            "[magenta]Saving hyperparameters summary...",
            spinner="bouncingBall",
        )
        pd.DataFrame(hyperparams).to_csv(HYPERPARAMETERS_SUMMARY_PATH)


if __name__ == "__main__":
    # clear_result_dir()
    create_result_dir()
    run_experiment()



from __future__ import annotations

import logging
import os
from pathlib import Path
from typing import TYPE_CHECKING, ClassVar

import numpy as np
import torch
import torch.nn as nn
from torch.utils.data import DataLoader

from climatrix.dataset.domain import Domain
from climatrix.decorators.runtime import log_input, raise_if_not_installed
from climatrix.reconstruct.base import BaseReconstructor
from climatrix.utils.hyperparameter import Hyperparameter

from .dataset import SIRENDataset
from .losses import (
    LossEntity,
    sdf_loss,
)
from .model import SIREN

if TYPE_CHECKING:
    from climatrix.dataset.base import BaseClimatrixDataset

log = logging.getLogger(__name__)


class SIRENReconstructor(BaseReconstructor):
    """
    A reconstructor that uses SIREN to reconstruct fields.

    SIREN (Sinusoidal Representation Networks) uses sinusoidal activation
    functions to learn continuous implicit neural representations of
    spatial fields from sparse observations.

    Parameters
    ----------
    dataset : BaseClimatrixDataset
        Source dataset to reconstruct from.
    target_domain : Domain
        Target domain to reconstruct onto.
    on_surface_points : int, default=1024
        Number of points to sample on the surface for training.
    hidden_features : int, default=256
        Number of features in each hidden layer.
    hidden_layers : int, default=4
        Number of hidden layers in the SIREN model.
    omega_0 : float, default=30.0
        Frequency multiplier for the first layer.
    omega_hidden : float, default=30.0
        Frequency multiplier for hidden layers.
    lr : float, default=1e-4
        Learning rate for the optimizer.
        Type: float, bounds: <unbounded>, default: 1e-3
    batch_size : int, default=256
        Batch size for training.
        Type: int, bounds: <unbounded>, default: 256
    num_epochs : int, default=100
        Number of epochs to train for.
        Type: int, bounds: <unbounded>, default: 5_000
    hidden_dim : int, default=256
        Hidden layer dimensions.
        Type: int, bounds: <unbounded>, default: 256
    num_layers : int, default=4
        Number of hidden layers.
        Type: int, bounds: <unbounded>, default: 4
    num_workers : int, default=0
        Number of worker processes for the dataloader.
    device : str, default="cuda"
        Device to run the model on ("cuda" or "cpu").
    gradient_clipping_value : float or None, default=None
        Value for gradient clipping (None to disable).
        Type: float, bounds: <unbounded>, default: 1.0
    checkpoint : str or os.PathLike or Path or None, default=None
        Path to save/load model checkpoint from.
    sdf_loss_weight : float, default=3000.0
        Weight for the SDF constraint loss.
    inter_loss_weight : float, default=100.0
        Weight for the interpolation consistency loss.
    normal_loss_weight : float, default=100.0
        Weight for the surface normal loss.
    grad_loss_weight : float, default=50.0
        Weight for the gradient regularization loss.

    Raises
    ------
    NotImplementedError
        If trying to use SIREN with a dynamic dataset.

    Notes
    -----
    Hyperparameters for optimization:
        - lr: float in (1e-5, 1e-2), default=1e-3
        - batch_size: int in (64, 1024), default=256
        - num_epochs: int in (100, 10_000), default=5_000
        - hidden_dim: int in (128, 512), default=256
        - num_layers: int in (3, 8), default=4
        - gradient_clipping_value: float in (0.1, 10.0), default=1.0
    """

    NAME: ClassVar[str] = "siren"

    # Hyperparameter descriptors
    lr = Hyperparameter(float, default=1e-3)
    batch_size = Hyperparameter(int, default=256)
    num_epochs = Hyperparameter(int, default=5_000)
    hidden_dim = Hyperparameter(int, default=256)
    num_layers = Hyperparameter(int, default=4)
    gradient_clipping_value = Hyperparameter(
        float, bounds=(0.1, 10.0), default=1.0
    )

    @log_input(log, level=logging.DEBUG)
    def __init__(
        self,
        dataset: BaseClimatrixDataset,
        target_domain: Domain,
        *,
        on_surface_points: int = 1024,
        hidden_features: int = 256,
        hidden_layers: int = 4,
        omega_0: float = 30.0,
        omega_hidden: float = 30.0,
        lr: float = 1e-4,
        num_epochs: int = 100,
        num_workers: int = 0,
        device: str = "cuda",
        gradient_clipping_value: float | None = None,
        checkpoint: str | os.PathLike | Path | None = None,
        sdf_loss_weight: float = 3e3,
        inter_loss_weight: float = 1e2,
        normal_loss_weight: float = 1e2,
        grad_loss_weight: float = 5e1,
    ) -> None:
        super().__init__(dataset, target_domain)

        if dataset.domain.is_dynamic:
            log.error("SIREN is not yet supported for dynamic datasets.")
            raise NotImplementedError(
                "SIREN is not yet supported for dynamic datasets."
            )

        if device == "cuda" and not torch.cuda.is_available():
            log.warning("CUDA is not available, falling back to CPU.")
            device = "cpu"

        self.on_surface_points = on_surface_points
        self.device = torch.device(device)
        self.hidden_features = hidden_features
        self.hidden_layers = hidden_layers
        self.omega_0 = omega_0
        self.omega_hidden = omega_hidden
        self.lr = lr
        self.num_epochs = num_epochs
        self.num_workers = num_workers
        self.gradient_clipping_value = gradient_clipping_value
        self.is_model_loaded = False
        self.checkpoint: Path | None = None
        self.sdf_loss_weight = sdf_loss_weight
        self.inter_loss_weight = inter_loss_weight
        self.normal_loss_weight = normal_loss_weight
        self.grad_loss_weight = grad_loss_weight

        input_coordinates_2d = dataset.domain.get_all_spatial_points()
        input_values_z = dataset.da.values.flatten().squeeze()

        self.train_dataset = SIRENDataset(
            input_coordinates_2d,
            input_values_z,
            on_surface_points=self.on_surface_points,
        )

        if checkpoint:
            self.checkpoint = Path(checkpoint).expanduser().absolute()
            log.info("Using checkpoint path: %s", self.checkpoint)

    def _init_model(self) -> nn.Module:
        """
        Initialize the 3D SIREN model.

        Returns
        -------
        nn.Module
            Initialized SIREN model on the appropriate device.
        """
        log.info("Initializing 3D SIREN model")
        return SIREN(
            in_features=3,
            out_features=1,
            hidden_features=self.hidden_features,
            hidden_layers=self.hidden_layers,
            outermost_linear=True,
            omega_0=self.omega_0,
            omega_hidden=self.omega_hidden,
        ).to(self.device)

    def _configure_optimizer(self, model: nn.Module) -> torch.optim.Optimizer:
        """
        Configure the optimizer for the model.

        Parameters
        ----------
        model : nn.Module
            The model to optimize.

        Returns
        -------
        torch.optim.Optimizer
            Configured Adam optimizer.
        """
        log.info(
            "Configuring Adam optimizer" " with learning rate: %0.6f", self.lr
        )
        optimizer = torch.optim.Adam(model.parameters(), lr=self.lr)
        return optimizer

    def _maybe_clip_grads(self, model: nn.Module) -> None:
        """
        Apply gradient clipping to model parameters if configured.

        Parameters
        ----------
        model : nn.Module
            The model whose gradients may need clipping.
        """
        if self.gradient_clipping_value:
            nn.utils.clip_grad_norm_(
                model.parameters(), self.gradient_clipping_value
            )

    @log_input(log, level=logging.DEBUG)
    def _apply_weights_to_loss_components(
        self, loss_component: LossEntity
    ) -> None:
        """
        Scales each loss component in the input LossEntity by its weight.

        Parameters
        ----------
        loss_component : LossEntity
            Contains sdf, inter, normal, and grad loss components.
        """
        loss_component.sdf *= self.sdf_loss_weight
        loss_component.inter *= self.inter_loss_weight
        loss_component.normal *= self.normal_loss_weight
        loss_component.grad *= self.grad_loss_weight

    def _maybe_load_checkpoint(
        self, model: nn.Module, checkpoint: Path | None
    ) -> nn.Module:
        """
        Load model weights from checkpoint if available.

        Parameters
        ----------
        model : nn.Module
            The model to load weights into.
        checkpoint : Path or None
            Path to the checkpoint file.

        Returns
        -------
        nn.Module
            The model, potentially with loaded weights.
        """
        if checkpoint and checkpoint.exists():
            log.info("Loading checkpoint from %s...", checkpoint)
            try:
                model.load_state_dict(
                    torch.load(checkpoint, map_location=self.device)
                )
                self.is_model_loaded = True
                log.info("Checkpoint loaded successfully.")
            except Exception as e:
                log.error("Error loading checkpoint: %s", e)
                log.info("Starting training from scratch.")
        else:
            log.info(
                "No checkpoint provided or checkpoint not found."
                " Starting training from scratch."
            )
        return model

    def _maybe_save_checkpoint(
        self, model: nn.Module, checkpoint: Path | None
    ) -> None:
        """
        Save model weights to checkpoint.

        Parameters
        ----------
        model : nn.Module
            The model to save weights from.
        checkpoint : Path or None
            Path to save the checkpoint to.
        """
        if checkpoint:
            if not checkpoint.parent.exists():
                log.info(
                    "Creating checkpoint directory:" " %s", checkpoint.parent
                )
                checkpoint.parent.mkdir(parents=True, exist_ok=True)
            log.info("Saving checkpoint to %s...", checkpoint)
            try:
                torch.save(model.state_dict(), checkpoint)
                log.info("Checkpoint saved successfully.")
            except Exception as e:
                log.error("Error saving checkpoint: %s", e)
        else:
            log.info(
                "Checkpoint saving skipped as no checkpoint path is provided."
            )

    def _normalize_coords(
        self, coords: np.ndarray, coord_min: np.ndarray, coord_max: np.ndarray
    ) -> np.ndarray:
        """
        Normalize coordinates to the range [-1, 1].

        Parameters
        ----------
        coords : np.ndarray
            Coordinates to normalize.
        coord_min : np.ndarray
            Minimum coordinate values.
        coord_max : np.ndarray
            Maximum coordinate values.

        Returns
        -------
        np.ndarray
            Normalized coordinates in range [-1, 1].
        """
        normalized = (coords - coord_min) / (coord_max - coord_min)
        normalized = normalized * 2.0 - 1.0
        return normalized

    def _denormalize_values(
        self, values: float | np.ndarray, min_value: float, max_value: float
    ) -> float | np.ndarray:
        """
        Denormalize values from [-1, 1] to original range.

        Parameters
        ----------
        values : float or np.ndarray
            Normalized values to denormalize.
        min_value : float
            Minimum value in original range.
        max_value : float
            Maximum value in original range.

        Returns
        -------
        float or np.ndarray
            Denormalized values in original range.
        """
        denormalized = (values + 1.0) / 2.0
        denormalized = denormalized * (max_value - min_value) + min_value
        return denormalized

    def _find_closest_value(
        self,
        decoder: nn.Module,
        x: float,
        y: float,
        z_min: float,
        z_max: float,
        num_samples: int = 1000,
    ) -> float:
        """
        Find the z value where the SDF is closest
        to zero for given x, y coordinates.

        Parameters
        ----------
        decoder : nn.Module
            The SIREN model to query.
        x : float
            Normalized x coordinate.
        y : float
            Normalized y coordinate.
        z_min : float
            Minimum normalized z value to sample.
        z_max : float
            Maximum normalized z value to sample.
        num_samples : int, default=1000
            Number of z values to sample.

        Returns
        -------
        float
            The z value where SDF is closest to zero.
        """
        z_values = torch.linspace(
            z_min, z_max, num_samples, device=self.device
        )

        x_values = torch.full((num_samples,), x, device=self.device)
        y_values = torch.full((num_samples,), y, device=self.device)
        points = torch.stack([x_values, y_values, z_values], dim=1)

        with torch.no_grad():
            sdf_values = decoder(points).squeeze()

        abs_sdf_values = torch.abs(sdf_values)
        min_idx = torch.argmin(abs_sdf_values)
        best_z = z_values[min_idx].item()
        return best_z

    @torch.no_grad()
    def _reconstruct_field(
        self, model: nn.Module, target_domain: Domain
    ) -> np.ndarray:
        """
        Reconstruct the field on the target domain
        using the trained SIREN model.

        Parameters
        ----------
        model : nn.Module
            Trained SIREN model.
        target_domain : Domain
            Target domain to reconstruct onto.

        Returns
        -------
        np.ndarray
            Reconstructed field values.
        """
        log.info("Querying 3D SIREN model on target domain...")

        target_lat_lon_2d = (
            target_domain.get_all_spatial_points()
        )  # shape: [N, 2]
        num_spatial_points = target_lat_lon_2d.shape[0]
        target_shape = target_domain.size

        xy_min = target_lat_lon_2d.min(axis=0)
        xy_max = target_lat_lon_2d.max(axis=0)

        coord_min = self.train_dataset.coord_min
        coord_max = self.train_dataset.coord_max

        height_min = coord_min[2].item()
        height_max = coord_max[2].item()
        height_margin = (height_max - height_min) * 0.2
        query_z_min = height_min - height_margin
        query_z_max = height_max + height_margin

        normalized_xy = self._normalize_coords(
            target_lat_lon_2d, xy_min, xy_max
        )

        z_values: list[float] = []
        for i in range(num_spatial_points):
            x_norm, y_norm = normalized_xy[i]

            best_z_norm = self._find_closest_value(
                model, x_norm.item(), y_norm.item(), -1, 1
            )

            best_z_real = self._denormalize_values(
                best_z_norm, query_z_min, query_z_max
            )
            z_values.append(best_z_real)

            if (i + 1) % 1000 == 0:
                log.info(f"Processed {i + 1}/{num_spatial_points} points")

        z_array = np.array(z_values).reshape(target_shape)
        return z_array

    @raise_if_not_installed("torch")
    def reconstruct(self) -> BaseClimatrixDataset:
        """
        Train (if necessary) and use a SIREN model to reconstruct the field.

        This method is the main entry point for using the SIREN reconstructor.
        It will train a new model if no checkpoint was loaded, and then use the
        model to reconstruct the field on the target domain.

        Returns
        -------
        BaseClimatrixDataset
            A dataset containing the reconstructed field.

        Raises
        ------
        ImportError
            If required dependencies are not installed.
        """
        from climatrix.dataset.base import BaseClimatrixDataset

        siren_model = self._init_model()
        siren_model = self._maybe_load_checkpoint(siren_model, self.checkpoint)

        if not self.is_model_loaded:
            log.info("Training 3D SIREN model (SDF loss)...")
            optimizer = self._configure_optimizer(siren_model)

            dataloader = DataLoader(
                self.train_dataset,
                batch_size=1,
                shuffle=True,
                num_workers=self.num_workers,
                pin_memory=True,
            )

            for epoch in range(1, self.num_epochs + 1):
                siren_model.train()
                epoch_loss = 0.0
                epoch_loss_components: dict[str, float] = {
                    "sdf": 0.0,
                    "inter": 0.0,
                    "normal_constraint": 0.0,
                    "grad_constraint": 0.0,
                }

                for coords_3d_normalized, gt_sdf, gt_normals in dataloader:
                    coords_3d_normalized.requires_grad_(True)

                    coords_3d_normalized = coords_3d_normalized.to(self.device)
                    gt_sdf = gt_sdf.to(self.device)
                    gt_normals = gt_normals.to(self.device)

                    pred_sdf = siren_model(coords_3d_normalized)

                    model_output: dict[str, torch.Tensor] = {
                        "model_in": coords_3d_normalized,
                        "model_out": pred_sdf,
                    }
                    gt_data: dict[str, torch.Tensor] = {
                        "sdf": gt_sdf,
                        "normals": gt_normals,
                    }
                    loss_component: LossEntity = sdf_loss(
                        model_output, gt_data
                    )

                    self._apply_weights_to_loss_components(loss_component)

                    train_loss = (
                        loss_component.sdf
                        + loss_component.inter
                        + loss_component.normal
                        + loss_component.grad
                    )

                    optimizer.zero_grad()
                    train_loss.backward()
                    self._maybe_clip_grads(siren_model)
                    optimizer.step()

                    epoch_loss += train_loss.item()
                    epoch_loss_components["sdf"] += loss_component.sdf.item()
                    epoch_loss_components[
                        "inter"
                    ] += loss_component.inter.item()
                    epoch_loss_components[
                        "normal_constraint"
                    ] += loss_component.normal.item()
                    epoch_loss_components[
                        "grad_constraint"
                    ] += loss_component.grad.item()

                if epoch % 100 == 0 or epoch == 1 or epoch == self.num_epochs:
                    avg_epoch_loss = epoch_loss / len(dataloader)
                    avg_epoch_components = {
                        k: v / len(dataloader)
                        for k, v in epoch_loss_components.items()
                    }
                    log.info(
                        f"Epoch {epoch}/{self.num_epochs}: "
                        f"Total Loss = {avg_epoch_loss:.6f}"
                    )
                    log.info(f" Loss Components: {avg_epoch_components}")

            log.info("Training finished.")
            self._maybe_save_checkpoint(siren_model, self.checkpoint)

        siren_model.eval()

        reconstructed_values = self._reconstruct_field(
            siren_model, self.target_domain
        )
        return BaseClimatrixDataset(
            self.target_domain.to_xarray(
                reconstructed_values, self.dataset.da.name
            )
        )



"""
Tests for the Domain.from_axes() builder pattern functionality.
"""

import numpy as np
import pytest
import xarray as xr

from climatrix.dataset.domain import Domain, DomainBuilder


class TestDomainBuilder:
    """Test cases for the DomainBuilder class and Domain.from_axes() method."""

    def test_domain_from_axes_creates_builder(self):
        """Test that Domain.from_axes() creates a DomainBuilder instance."""
        builder = Domain.from_axes()
        assert isinstance(builder, DomainBuilder)

    def test_basic_sparse_domain_creation(self):
        """Test creating a basic sparse domain with lat/lon."""
        domain = (
            Domain.from_axes()
            .lat(latitude=[1, 2, 3])
            .lon(longitude=[4, 5, 6])
            .sparse()
        )
        assert domain.is_sparse
        assert len(domain.latitude.values) == 3
        assert len(domain.longitude.values) == 3
        np.testing.assert_array_equal(domain.latitude.values, [1, 2, 3])
        np.testing.assert_array_equal(domain.longitude.values, [4, 5, 6])

    def test_basic_dense_domain_creation(self):
        """Test creating a basic dense domain with lat/lon."""
        domain = (
            Domain.from_axes()
            .lat(latitude=[1, 2])
            .lon(longitude=[3, 4])
            .dense()
        )
        assert not domain.is_sparse
        assert len(domain.latitude.values) == 2
        assert len(domain.longitude.values) == 2
        assert domain.shape == (2, 2)

    def test_vertical_axis_in_sparse_domain(self):
        """Test adding vertical axis to sparse domain."""
        domain = (
            Domain.from_axes()
            .vertical(depth=slice(10, 50, 10))
            .lat(latitude=[1, 2])
            .lon(longitude=[3, 4])
            .sparse()
        )

        assert domain.has_axis("vertical")
        assert domain.vertical.name == "depth"
        np.testing.assert_array_equal(domain.vertical.values, [10, 20, 30, 40])

    def test_time_axis_in_domain(self):
        """Test adding time axis to domain."""
        domain = (
            Domain.from_axes()
            .lat(latitude=[1, 2])
            .lon(longitude=[3, 4])
            .time(time=["2020-01-01", "2020-01-02"])
            .dense()
        )

        assert domain.has_axis("time")
        assert domain.time.name == "time"

    def test_slice_to_array_conversion(self):
        """Test that slices are properly converted to arrays."""
        builder = DomainBuilder()
        result = builder._convert_slice_to_array(slice(1, 10, 2))
        expected = np.array([1, 3, 5, 7, 9])
        np.testing.assert_array_equal(result, expected)

    def test_custom_axis_names(self):
        """Test using custom names for axes."""
        domain = (
            Domain.from_axes()
            .vertical(pressure=[1000, 850, 500])
            .lat(latitude=[0, 1])
            .lon(longitude=[2, 3])
            .sparse()
        )

        assert domain.vertical.name == "pressure"
        np.testing.assert_array_equal(domain.vertical.values, [1000, 850, 500])

    def test_method_chaining(self):
        """Test that methods can be chained."""
        builder = (
            Domain.from_axes()
            .lat(latitude=[1])
            .lon(longitude=[2])
            .vertical(depth=[10])
            .time(time=["2020-01-01"])
        )

        assert len(builder._axes_config) == 4

    def test_sparse_validation_mismatched_coordinates(self):
        """Test validation for sparse domains with mismatched coordinate lengths."""
        with pytest.raises(
            ValueError,
            match="latitude and longitude arrays must have the same length",
        ):
            Domain.from_axes().lat(latitude=[1, 2]).lon(
                longitude=[3, 4, 5]
            ).sparse()

    def test_sparse_validation_missing_latitude(self):
        """Test validation for sparse domains missing latitude."""
        with pytest.raises(ValueError, match="Latitude axis is required"):
            Domain.from_axes().lon(longitude=[1, 2]).sparse()

    def test_sparse_validation_missing_longitude(self):
        """Test validation for sparse domains missing longitude."""
        with pytest.raises(ValueError, match="Longitude axis is required"):
            Domain.from_axes().lat(latitude=[1, 2]).sparse()

    def test_multiple_kwargs_error(self):
        """Test that providing multiple kwargs to a method raises an error."""
        with pytest.raises(
            ValueError, match="Exactly one keyword argument must be provided"
        ):
            Domain.from_axes().vertical(depth=[1], pressure=[2])

    def test_slice_without_step_error(self):
        """Test that slices without step raise an error."""
        with pytest.raises(ValueError, match="Slice step must be specified"):
            Domain.from_axes().lat(latitude=slice(1, 10)).lon(
                longitude=[1]
            ).sparse()

    def test_equivalence_with_from_lat_lon(self):
        """Test that the new method produces equivalent results to from_lat_lon."""
        lat_vals = [1, 2, 3]
        lon_vals = [4, 5, 6]

        # Create using old method
        domain_old = Domain.from_lat_lon(
            lat=np.array(lat_vals), lon=np.array(lon_vals), kind="sparse"
        )

        # Create using new method
        domain_new = (
            Domain.from_axes().lat(lat=lat_vals).lon(lon=lon_vals).sparse()
        )

        assert domain_old.is_sparse == domain_new.is_sparse
        np.testing.assert_array_equal(
            domain_old.latitude.values, domain_new.latitude.values
        )
        np.testing.assert_array_equal(
            domain_old.longitude.values, domain_new.longitude.values
        )

    def test_full_example_from_issue(self):
        """Test the exact example from the issue description."""
        # Domain.from_axes().vertical(depth=slice(10, 100, 1)).lat(latitude=[1,2,3,4]).lon(...).time(....).sparse()
        domain = (
            Domain.from_axes()
            .vertical(depth=slice(10, 100, 30))
            .lat(latitude=[1, 2, 3, 4])
            .lon(longitude=[5, 6, 7, 8])
            .time(time=["2020-01-01", "2020-01-02"])
            .sparse()
        )

        assert domain.is_sparse
        assert domain.has_axis("vertical")
        assert domain.has_axis("time")
        assert domain.vertical.name == "depth"
        assert domain.time.name == "time"
        assert len(domain.latitude.values) == 4
        assert len(domain.longitude.values) == 4



"""Tests for Bayesian hyperparameter optimization."""

from datetime import datetime
from unittest.mock import MagicMock
import importlib

import numpy as np
import pytest
import xarray as xr

from climatrix import BaseClimatrixDataset
from climatrix.optim.bayesian import HParamFinder, MetricType
from climatrix.reconstruct.base import BaseReconstructor
from tests.unit.test_utils import skip_on_error


class TestMetricType:
    """Test the MetricType enum."""

    def test_metric_values(self):
        """Test metric enum values."""
        assert MetricType.MAE == "mae"
        assert MetricType.MSE == "mse"
        assert MetricType.RMSE == "rmse"


class TestBaseReconstructorRegistry:
    """Test the BaseReconstructor registry system."""

    def test_idw_class(self):
        """Test getting IDW reconstruction class."""
        cls = BaseReconstructor.get("idw")
        assert cls.__name__ == "IDWReconstructor"

    def test_ok_class(self):
        """Test getting Ordinary Kriging reconstruction class."""
        cls = BaseReconstructor.get("ok")
        assert cls.__name__ == "OrdinaryKrigingReconstructor"

    def test_case_insensitive(self):
        """Test that method names are case insensitive."""
        cls_lower = BaseReconstructor.get("idw")
        cls_upper = BaseReconstructor.get("IDW")
        cls_mixed = BaseReconstructor.get("IdW")

        assert cls_lower == cls_upper == cls_mixed

    def test_unknown_method(self):
        """Test error for unknown method."""
        with pytest.raises(ValueError, match="Unknown method"):
            BaseReconstructor.get("unknown_method")


class TestHyperparameterProperty:
    """Test the hparams classmethod system."""

    def test_idw_hparams(self):
        """Test IDW hyperparameters."""
        from climatrix.reconstruct.idw import IDWReconstructor

        hparams = IDWReconstructor.get_hparams()

        assert "power" in hparams
        assert "k" in hparams
        assert "k_min" in hparams

        assert issubclass(hparams["power"]["type"], float)
        assert issubclass(hparams["k"]["type"], int)
        # NOTE: power is unbounded
        assert "bounds" not in hparams["power"]


class TestHParamFinder:
    """Test the HParamFinder class."""

    @pytest.fixture
    def sparse_dataset(self):
        """Create a sparse dataset for testing."""
        return BaseClimatrixDataset(
            xr.DataArray(
                data=np.random.rand(5, 1),
                dims=("point", "time"),
                coords={
                    "point": np.arange(5),
                    "time": np.array(
                        [datetime(2000, 1, 1)], dtype="datetime64"
                    ),
                    "latitude": (("point",), np.array([-90, -45, 0, 45, 90])),
                    "longitude": (
                        ("point",),
                        np.array([-180, -90, 0, 90, 180]),
                    ),
                },
            )
        )

    @pytest.fixture
    def dense_dataset(self):
        """Create a dense dataset for testing."""
        return BaseClimatrixDataset(
            xr.DataArray(
                data=np.random.rand(1, 3, 3),
                dims=("time", "latitude", "longitude"),
                coords={
                    "time": np.array(
                        [datetime(2000, 1, 1)], dtype="datetime64"
                    ),
                    "latitude": (("latitude",), np.array([-45, 0, 45])),
                    "longitude": (("longitude",), np.array([-90, 0, 90])),
                },
            )
        )

    def test_init_basic(self, sparse_dataset, dense_dataset):
        """Test basic initialization."""
        finder = HParamFinder("idw", sparse_dataset, dense_dataset)

        assert finder.train_dset is sparse_dataset
        assert finder.val_dset is dense_dataset
        assert finder.metric == MetricType.MAE
        assert finder.method == "idw"
        assert finder.random_seed == 42
        assert finder.n_init_points + finder.n_iter == 100
        assert finder.bounds is not None

    def test_init_with_parameters(self, sparse_dataset, dense_dataset):
        """Test initialization with custom parameters."""
        finder = HParamFinder(
            "ok",
            sparse_dataset,
            dense_dataset,
            metric="mse",
            explore=0.5,
            n_iters=50,
            random_seed=123,
        )

        assert finder.metric == MetricType.MSE
        assert finder.method == "ok"
        assert finder.random_seed == 123
        assert finder.n_init_points == 25  # 50 * 0.5
        assert finder.n_iter == 25

    def test_include_parameters(self, sparse_dataset, dense_dataset):
        """Test parameter inclusion."""
        finder = HParamFinder(
            "idw", sparse_dataset, dense_dataset, include=["power", "k"]
        )

        assert set(finder.bounds.keys()) == {"power", "k"}

    def test_exclude_parameters(self, sparse_dataset, dense_dataset):
        """Test parameter exclusion."""
        finder = HParamFinder(
            "idw", sparse_dataset, dense_dataset, exclude="k"
        )

        expected_params = {"power", "k_min"}
        assert set(finder.bounds.keys()) == expected_params

    def test_include_exclude_both(self, sparse_dataset, dense_dataset):
        """Test that include and exclude can be used together if no common keys."""
        finder = HParamFinder(
            "idw",
            sparse_dataset,
            dense_dataset,
            include=["power", "k"],
            exclude=["k_min"],
        )

        assert set(finder.bounds.keys()) == {"power", "k"}

    def test_include_exclude_common_keys(self, sparse_dataset, dense_dataset):
        """Test that include and exclude cannot have common keys."""
        with pytest.raises(
            ValueError,
            match="Cannot specify same parameters in both include and exclude",
        ):
            HParamFinder(
                "idw",
                sparse_dataset,
                dense_dataset,
                include=["power", "k"],
                exclude=["k"],
            )

    def test_custom_bounds_override(self, sparse_dataset, dense_dataset):
        """Test custom bounds override."""
        custom_bounds = {"power": (100.0, 330.0), "k": (20, 89)}
        finder = HParamFinder(
            "idw", sparse_dataset, dense_dataset, bounds=custom_bounds
        )
        assert finder.bounds["power"] == (100.0, 330.0, float)
        assert finder.bounds["k"] == (20, 89, int)

    def test_invalid_explore(self, sparse_dataset, dense_dataset):
        """Test invalid explore parameter."""
        with pytest.raises(ValueError, match="explore must be in the range"):
            HParamFinder("idw", sparse_dataset, dense_dataset, explore=0.0)

        with pytest.raises(ValueError, match="explore must be in the range"):
            HParamFinder("idw", sparse_dataset, dense_dataset, explore=1.0)

    def test_invalid_n_iters(self, sparse_dataset, dense_dataset):
        """Test invalid n_iters parameter."""
        with pytest.raises(ValueError, match="n_iters must be >= 1"):
            HParamFinder("idw", sparse_dataset, dense_dataset, n_iters=0)

    def test_invalid_metric(self, sparse_dataset, dense_dataset):
        """Test invalid metric parameter."""
        with pytest.raises(
            ValueError, match="'invalid_metric' is not a valid MetricType"
        ):
            HParamFinder(
                "idw", sparse_dataset, dense_dataset, metric="invalid_metric"
            )

    def test_invalid_dataset_types(self, sparse_dataset):
        """Test invalid dataset types."""
        with pytest.raises(
            TypeError, match="train_dset must be a BaseClimatrixDataset"
        ):
            HParamFinder("idw", "not_a_dataset", sparse_dataset)

        with pytest.raises(
            TypeError, match="val_dset must be a BaseClimatrixDataset"
        ):
            HParamFinder("idw", sparse_dataset, "not_a_dataset")

    @pytest.mark.skipif(not importlib.util.find_spec("optuna"), reason="`optuna` package is not installed")
    def test_evaluate_params(self, sparse_dataset, dense_dataset):
        """Test parameter evaluation (without full optimization)."""
        from optuna import Trial
        finder = HParamFinder(
            "idw",
            sparse_dataset,
            dense_dataset,
        )
        trial = MagicMock(spec=Trial)
        trial.suggest_int.side_effect = [2, 5]
        trial.suggest_float.side_effect = [5.0]
        
        result = finder._evaluate_params(trial)
        assert isinstance(result, float)
        assert result <= 0

    @skip_on_error(ImportError)
    def test_optimize_mock(self, sparse_dataset, dense_dataset):
        _ = HParamFinder(
            "idw", sparse_dataset, dense_dataset, n_iters=5, include=["power"]
        ).optimize()

    def test_optimize_on_explor_high(self, sparse_dataset, dense_dataset):
        finder = HParamFinder(
            "idw",
            sparse_dataset,
            dense_dataset,
            n_iters=10,
            explore=0.99,  # High exploration
        )
        finder.optimize()



from functools import wraps

import pytest


def skip_on_error(*err_types):
    """
    Decorator to skip a test if any of the specified error types are raised.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except err_types as e:
                pytest.skip(f"Skipped due to error: {e}")

        return wrapper

    return decorator



import xarray as xr

import climatrix as cm

dset = xr.open_dataset(
    "/home/jakub/tul/research/climatrix/data/static-era5-land.nc"
)
europe = dset.cm.itime(0)
sample = (
    europe.sample_uniform(number=100_000, nan="resample")
    .subset(north=71, south=36, west=-24, east=35)
    .plot()
)



from __future__ import annotations

import importlib.resources
import logging
from collections import namedtuple
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
import torch
from scipy.spatial import cKDTree
from sklearn.preprocessing import MinMaxScaler
from torch.utils.data import Dataset

from climatrix.decorators.runtime import log_input

if TYPE_CHECKING:
    pass

SdfEntry = namedtuple("SdfEntry", ["coordinates", "normals", "sdf"])

log = logging.getLogger(__name__)

_ELEVATION_DATASET_PATH: Path = importlib.resources.files(
    "climatrix.reconstruct.sinet"
).joinpath("resources", "lat_lon_elevation.npy")


def load_elevation_dataset() -> np.ndarray:
    """
    Load the elevation dataset (download if needed).

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        Tuple containing the coordinates and elevation data as numpy arrays.
        The first array contains the coordinates (latitude, longitude),
        and the second array contains the elevation values.
    """
    log.debug("Loading elevation dataset...")
    try:
        data = np.load(_ELEVATION_DATASET_PATH)
    except FileNotFoundError:
        log.info("Elevation dataset not found, downloading...")
        # TODO: download to be implemented
        raise NotImplementedError
        log.info("Elevation dataset downloaded successfully.")
        data = np.load(_ELEVATION_DATASET_PATH)
    return data[:, :-1], MinMaxScaler((-1, 1)).fit_transform(
        data[:, -1].reshape(-1, 1)
    )


def query_features(
    tree: cKDTree, values: np.ndarray, query_points: np.ndarray
):
    log.debug("Querying nearest neighbours...")
    distances, indices = tree.query(query_points, k=1)
    if np.any(distances > 0.1):
        log.warning(
            "Some coordinates are too far from the known data points. The maximum distance is %f.",
            distances.max(),
        )
    return values[indices]


class SiNETDatasetGenerator:
    train_coordinates: np.ndarray
    train_field: np.ndarray
    target_coordinates: np.ndarray | None
    field_transformer: MinMaxScaler
    radius: float

    @log_input(log, level=logging.DEBUG)
    def __init__(
        self,
        spatial_points: np.ndarray,
        field: np.ndarray,
        *,
        target_coordinates: np.ndarray | None = None,
        degree: bool = True,
        radius: float = 1.0,
        val_portion: float | None = None,
        validation_coordinates: np.ndarray | None = None,
        validation_field: np.ndarray | None = None,
        use_elevation: bool = False,
    ) -> None:
        """
        Initialize a SiNET dataset generator.

        Parameters
        ----------
        spatial_points : np.ndarray
            Array on shape Nx2 with latitudes and longitudes of training points.
        field : np.ndarray
            Array of shape (N,) with field values at training points.
        target_coordinates : np.ndarray | None, optional
            Array on shape Nx2 with latitudes and longitudes of target points.
        degree : bool, optional
            Whether the input latitudes and longitudes are in degrees.
            Defaults to True.
        radius : float, optional
            The radius of the sphere. Defaults to 1.0.
        val_portion : float | None, optional
            Portion of the training data to use for validation. Defaults to 0.2.
        validation_coordinates : np.ndarray | None, optional
            Array on shape Nx2 with latitudes and longitudes of validation points.
            Cannot be used together with `val_portion`.
            Must be provided if `validation_field` is given.
        validation_field : np.ndarray | None, optional
            Array of shape (N,) with field values at validation points.
            Cannot be used together with `val_portion`.
            Must be provided if `validation_coordinates` is given.
        use_elevation: bool, optional
            Whether to use elevation data. Defaults to False.
        """
        self.field_transformer = MinMaxScaler((0, 1))
        field = self.field_transformer.fit_transform(field.reshape(-1, 1))
        self.target_coordinates = target_coordinates
        if degree:
            log.debug("Converting degrees to radians...")
            spatial_points = np.deg2rad(spatial_points)
            if target_coordinates is not None:
                self.target_coordinates = np.deg2rad(target_coordinates)

        if val_portion is not None:
            if not (0 < val_portion < 1):
                log.error(
                    "Validation portion must be between 0 and 1, got %f.",
                    val_portion,
                )
                raise ValueError(
                    "Validation portion must be between 0 and 1, got %f."
                    % val_portion
                )
            if (
                validation_coordinates is not None
                or validation_field is not None
            ):
                log.error(
                    "Cannot use both `val_portion` and `validation_coordinates`/`validation_field`."
                )
                raise ValueError(
                    "Cannot use both `val_portion` and `validation_coordinates`/`validation_field`."
                )
            log.debug("Splitting train and validation datasets...")
            (
                self.train_coordinates,
                self.train_field,
                self.val_coordinates,
                self.val_field,
            ) = self._split_train_val(spatial_points, field, val_portion)
        else:
            if validation_coordinates is None or validation_field is None:
                log.error(
                    "Validation coordinates and field must be provided if `val_portion` is not used."
                )
                raise ValueError(
                    "Validation coordinates and field must be provided if `val_portion` is not used."
                )
            log.debug("Using provided validation dataset...")
            self.train_coordinates = spatial_points
            self.train_field = field
            self.val_coordinates = validation_coordinates
            self.val_field = validation_field

        ckdtree = None
        self.radius = radius
        if use_elevation:
            coords, self.elevation = load_elevation_dataset()
            ckdtree = cKDTree(np.deg2rad(coords))
            self._extend_input_features(ckdtree, self.elevation)

        # # Plot europe lat/lon
        # import cartopy.crs as ccrs
        # import cartopy.feature as cfeature
        # import matplotlib.pyplot as plt

        # fig = plt.figure(figsize=(10, 8))
        # ax = fig.add_subplot(1, 1, 1, projection=ccrs.PlateCarree())
        # scatter = ax.scatter(
        #     np.rad2deg(self.target_coordinates[:, 1]),
        #     np.rad2deg(self.target_coordinates[:, 0]),
        #     c=self.target_coordinates[:, 2],
        #     cmap="viridis",
        #     s=50,
        #     transform=ccrs.PlateCarree(),
        #     edgecolor="k",
        #     linewidth=0.5,
        #     alpha=0.7,
        # )
        # ax.add_feature(cfeature.BORDERS, linestyle=":", edgecolor="gray")
        # ax.add_feature(cfeature.COASTLINE, linewidth=0.8)
        # # ax.set_extent([0, 40, 30, 60], crs=ccrs.PlateCarree())
        # gl = ax.gridlines(
        #     crs=ccrs.PlateCarree(),
        #     draw_labels=True,
        #     linewidth=0.5,
        #     color="gray",
        #     alpha=0.5,
        #     linestyle="--",
        # )
        # gl.top_labels = False
        # gl.right_labels = False

        # # Set title
        # ax.set_title("Europe Latitude/Longitude")

        # plt.show()

    @property
    def n_features(self) -> int:
        """
        Number of features in the dataset.

        Returns
        -------
        int
            Number of features.
        """
        return self.train_coordinates.shape[1]

    def _extend_input_features(
        self, tree: cKDTree, values: np.ndarray
    ) -> np.ndarray:
        train_extra_feature = query_features(
            tree, values, self.train_coordinates[:, :2]
        )
        self.train_coordinates = np.concatenate(
            [self.train_coordinates, train_extra_feature.reshape(-1, 1)],
            axis=1,
        )

        val_extra_feature = query_features(
            tree, values, self.val_coordinates[:, :2]
        )
        self.val_coordinates = np.concatenate(
            [self.val_coordinates, val_extra_feature.reshape(-1, 1)], axis=1
        )

        if self.target_coordinates is None:
            return
        target_extra_feature = query_features(
            tree, values, self.target_coordinates[:, :2]
        )
        self.target_coordinates = np.concatenate(
            [self.target_coordinates, target_extra_feature.reshape(-1, 1)],
            axis=1,
        )

    def _split_train_val(
        self, coordinates: np.ndarray, field: np.ndarray, val_portion: float
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        log.debug("Splitting train and validation datasets...")
        num_samples = coordinates.shape[0]
        indices = np.arange(num_samples)
        np.random.seed(0)
        np.random.shuffle(indices)
        split_index = int(num_samples * (1 - val_portion))
        train_indices = indices[:split_index]
        val_indices = indices[split_index:]
        return (
            coordinates[train_indices],
            field[train_indices],
            coordinates[val_indices],
            field[val_indices],
        )

    @staticmethod
    def convert_spherical_to_cartesian(
        coordinates: np.ndarray, radius: float = 1.0
    ) -> np.ndarray:
        log.debug("Converting coordinates to cartesian...")
        x = radius * np.cos(coordinates[:, 0]) * np.cos(coordinates[:, 1])
        y = radius * np.cos(coordinates[:, 0]) * np.sin(coordinates[:, 1])
        z = radius * np.sin(coordinates[:, 0])
        return np.stack((x, y, z), axis=1)

    @property
    def train_dataset(self) -> Dataset:
        return torch.utils.data.TensorDataset(
            torch.from_numpy(self.train_coordinates).float(),
            torch.from_numpy(self.train_field).float(),
        )

    @property
    def val_dataset(self) -> Dataset:
        return torch.utils.data.TensorDataset(
            torch.from_numpy(self.val_coordinates).float(),
            torch.from_numpy(self.val_field).float(),
        )

    @property
    def target_dataset(self) -> Dataset:
        if self.target_coordinates is None:
            raise ValueError("Target coordinates are not set.")
        return torch.utils.data.TensorDataset(
            torch.from_numpy(self.target_coordinates).float()
        )



from __future__ import annotations

import logging

import numpy as np
import torch
from sklearn.decomposition import PCA
from sklearn.neighbors import NearestNeighbors
from torch.utils.data import Dataset

log = logging.getLogger(__name__)


class SIRENDataset(Dataset):
    """
    Dataset for training SIREN on 3D point clouds.

    This class prepares data for implicit neural representation
    learning by processing coordinate-value pairs and creating on-surface
    and off-surface points. The dataset normalizes coordinates and calculates
    surface normals for better learning.
    """

    def __init__(
        self,
        coordinates: np.ndarray,
        values: np.ndarray,
        on_surface_points: int,
    ) -> None:
        """
        Initialize the SIREN dataset with coordinates and corresponding values.

        Parameters
        ----------
        coordinates : np.ndarray
            Array of shape [N, 2] containing the coordinate points
        values : np.ndarray
            Array of shape [N] containing the values at each coordinate
        on_surface_points : int
            Number of on-surface points to sample per batch

        Raises
        ------
        ValueError
            If the number of coordinates doesn't match the number of values
        """
        if len(values.shape) == 1:
            values = values.reshape(-1, 1)
        if coordinates.shape[0] != values.shape[0]:
            log.error(
                f"Mismatch between coordinates ({coordinates.shape[0]})"
                f" and values ({values.shape[0]}) count"
            )
            raise ValueError(
                f"Mismatch between coordinates ({coordinates.shape[0]})"
                f" and values ({values.shape[0]}) count"
            )

        if coordinates.shape[0] < on_surface_points:
            log.warning(
                f"Only {coordinates.shape[0]} points available, "
                f"but requested {on_surface_points} on-surface points. "
                f"Reducing on_surface_points to match available data."
            )
            self.on_surface_points = coordinates.shape[0]
        else:
            self.on_surface_points = on_surface_points

        points_3d = np.concatenate([coordinates, values], axis=1)
        log.info(
            f"Created 3D points from coordinates and values: {points_3d.shape}"
        )

        self.coord_min = np.min(points_3d, axis=0)
        self.coord_max = np.max(points_3d, axis=0)

        normals_np = self._calculate_normals(points_3d)

        self.points_3d = torch.tensor(points_3d, dtype=torch.float32)
        self.normals = torch.tensor(normals_np, dtype=torch.float32)

        if self.normals.shape != self.points_3d.shape:
            log.error(
                f"Mismatch between points ({self.points_3d.shape})"
                f" and calculated normals ({self.normals.shape})"
                f" shape. Setting normals to zeros."
            )
            self.normals = torch.zeros_like(self.points_3d)

        self.points_3d = self._normalize_coords(self.points_3d)
        self.points_3d = torch.tensor(self.points_3d, dtype=torch.float32)
        log.info("Normalized 3D on-surface points.")

        self.total_points = self.points_3d.shape[0]

        log.info(
            f"Created dataset with {self.on_surface_points}"
            f" total on-surface points"
        )

    def _calculate_normals(self, points: np.ndarray) -> np.ndarray:
        """
        Calculate surface normals from point cloud data.

        Parameters
        ----------
        points : np.ndarray
            Array of shape [N, 3] containing 3D points

        Returns
        -------
        np.ndarray
            Array of shape [N, 3] containing the normal vectors
        """
        k_neighbors = min(30, points.shape[0] - 1)

        nnbrs = NearestNeighbors(
            n_neighbors=k_neighbors + 1, algorithm="auto"
        ).fit(points)
        distances, indices = nnbrs.kneighbors(points)

        normals = np.zeros_like(points)
        for i in range(points.shape[0]):
            neighbors = points[indices[i, 1:]]

            neighbors = neighbors - np.mean(neighbors, axis=0)

            pca = PCA(n_components=3)
            pca.fit(neighbors)

            normals[i] = pca.components_[2]

        norms = np.linalg.norm(normals, axis=1, keepdims=True)
        normals = normals / np.maximum(norms, 1e-10)

        return normals

    def _normalize_coords(self, coords_3d: torch.Tensor) -> torch.Tensor:
        """
        Normalize coordinates to the range [-1, 1].

        Parameters
        ----------
        coords_3d : torch.Tensor
            Containing 3D coordinates

        Returns
        -------
        torch.Tensor
            Containing normalized coordinates
        """
        normalized = (coords_3d - self.coord_min) / (
            self.coord_max - self.coord_min
        )
        normalized = normalized * 2.0 - 1.0
        return normalized

    def __len__(self) -> int:
        """
        Return the number of batches in the dataset.

        Returns
        -------
        int
            Number of available batches
        """
        return self.total_points // self.on_surface_points

    def __getitem__(
        self, idx: int
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Return a batch of mixed on-surface and off-surface points.

        Parameters
        ----------
        idx : int
            Index of the batch to retrieve (not used in this
            implementation as points are randomly sampled)

        Returns
        -------
        tuple
            Contains:
            - coords : torch.Tensor
                Containing coordinates
            - sdf : torch.Tensor
                Containing signed distance values
            - normals : torch.Tensor
                Containing normal vectors
        """
        on_surface_samples = self.on_surface_points
        off_surface_samples = self.on_surface_points

        rand_idcs = torch.randint(0, self.total_points, (on_surface_samples,))
        on_surface_coords = self.points_3d[rand_idcs]
        on_surface_normals = self.normals[rand_idcs]

        off_surface_coords = torch.FloatTensor(
            off_surface_samples, 3
        ).uniform_(-1, 1)

        off_surface_normals = (
            torch.ones((off_surface_samples, 3), dtype=torch.float32) * -1
        )

        on_surface_sdf = torch.zeros(
            (on_surface_samples, 1), dtype=torch.float32
        )
        off_surface_sdf = (
            torch.ones((off_surface_samples, 1), dtype=torch.float32) * -1
        )

        coords = torch.cat([on_surface_coords, off_surface_coords], dim=0)
        normals = torch.cat([on_surface_normals, off_surface_normals], dim=0)
        sdf = torch.cat([on_surface_sdf, off_surface_sdf], dim=0)

        shuffle_idx = torch.randperm(coords.shape[0])

        return (
            coords[shuffle_idx],
            sdf[shuffle_idx],
            normals[shuffle_idx],
        )



from typing import Callable

import torch


class EarlyStopping:
    step: Callable[[float, torch.nn.Module | None], bool]

    def __init__(
        self,
        patience: int | None,
        delta=0.0,
        checkpoint_path=None,
        verbose=False,
    ):
        self.patience = patience
        self.delta = delta
        self.checkpoint_path = checkpoint_path
        self.verbose = verbose

        self.counter = 0
        self.best_score = None
        self.early_stop = False
        self.best_val_metric = float("inf")

        if self.patience is None:
            self.step = self.dummy_call
        else:
            self.step = self.call

    def dummy_call(self, val_metric, model=None):
        return False

    def call(self, val_metric, model=None):
        score = -val_metric

        if self.best_score is None:
            self.best_score = score
            if self.checkpoint_path and model:
                self.save_checkpoint(val_metric, model)
        elif score < self.best_score + self.delta:
            self.counter += 1
            if self.verbose:
                print(
                    f"[EarlyStopping] No improvement. Counter: {self.counter}/{self.patience}"
                )
            if self.counter >= self.patience:
                self.early_stop = True
        else:
            self.best_score = score
            if self.checkpoint_path and model:
                self.save_checkpoint(val_metric, model)
            self.counter = 0

        return self.early_stop

    def save_checkpoint(self, val_metric, model):
        """Save the model state_dict to the checkpoint path."""
        if self.verbose:
            print(
                f"[EarlyStopping] Validation metric improved "
                f"({self.best_val_metric:.6f} -> {val_metric:.6f}). Saving model."
            )
        torch.save(model.state_dict(), self.checkpoint_path)
        self.best_val_metric = val_metric



import functools
import importlib
import logging
from typing import Callable


def raise_if_not_installed(
    *packages, custom_error: Exception | None = None
) -> Callable:
    """
    Decorator to raise an ImportError if a package is not installed.
    """

    def decorator(func):

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            missing = []
            for pkg in packages:
                if importlib.util.find_spec(pkg) is None:
                    missing.append(pkg)
            if missing:
                if custom_error:
                    raise custom_error
                raise ImportError(
                    "The following packages are required but not "
                    f"installed: {', '.join(missing)}. "
                    "Please install them using pip or conda before "
                    f"calling '{func.__name__}()'."
                )
            return func(*args, **kwargs)

        return wrapper

    return decorator


def log_input(log: logging.Logger, level: int = logging.INFO):
    """Decorator to facilitte logging the input of a function."""

    def decorator(func):

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            log.log(level, f"Input: {args}, {kwargs}")
            return func(*args, **kwargs)

        return wrapper

    return decorator



import logging
from enum import Enum

log = logging.getLogger(__name__)

from climatrix.reconstruct.idw import IDWReconstructor

reconstructors = {
    "IDW": IDWReconstructor,
}

try:
    from climatrix.reconstruct.kriging import OrdinaryKrigingReconstructor

    reconstructors["OK"] = OrdinaryKrigingReconstructor
except ImportError:
    log.warning(
        "OrdinaryKrigingReconstructor not available. "
        "Install climatrix[ok] to use it."
    )
    pass

try:
    from climatrix.reconstruct.sinet.sinet import SiNETReconstructor

    reconstructors["SINET"] = SiNETReconstructor
except ImportError:
    pass
try:
    from climatrix.reconstruct.siren.siren import SIRENReconstructor

    reconstructors["SIREN"] = SIRENReconstructor
except ImportError:
    log.warning(
        "SIRENReconstructor not available. " "Install climatrix[ml] to use it."
    )
    pass

ReconstructionType = Enum(
    "ReconstructionType",
    {name: cls for name, cls in reconstructors.items()},
)


def __missing__(cls, value):
    raise ValueError(f"Unknown reconstruction method: {value}")


def get(cls, value: str | ReconstructionType):
    """
    Get the reconstruction type given by `value`.

    If `value` is an instance of ReconstructionType,
    return it as is.
    If `value` is a string, return the corresponding
    ReconstructionType.
    If `value` is neither an instance of ReconstructionType
    nor a string,
    raise a ValueError.

    Parameters
    ----------
    value : str | ReconstructionType
        The reconstruction type to get.

    Returns
    -------
    ReconstructionType
        The reconstruction type.

    Raises
    ------
    ValueError
        If `value` is not a valid reconstruction type.
    """
    if isinstance(value, cls):
        return value
    if not isinstance(value, str):
        raise TypeError(
            f"Invalid reconstruction type: {value!r}. "
            "Expected a string or an instance of ReconstructionType."
        )
    try:
        return cls[value.upper()]
    except KeyError:
        raise ValueError(
            f"Unknown reconstruction type: {value}. "
            f"Supported types are: ({', '.join(cls.list())})."
            "Ensure that the required packages are installed."
        )


def list(cls) -> list[str]:
    """
    List all available reconstruction types.

    Returns
    -------
    list[str]
        A list of all available reconstruction types.
    """
    return [name.lower() for name in cls.__members__.keys()]


setattr(ReconstructionType, "__missing__", classmethod(__missing__))
setattr(ReconstructionType, "get", classmethod(get))
setattr(ReconstructionType, "list", classmethod(list))



import functools
import inspect
from numbers import Number


def _assert_is_binary_func(func):
    if not callable(func):
        raise TypeError(
            f"Expected a function, but found: {type(func).__name__}"
        )
    arguments = (
        inspect.getfullargspec(func).args
        + inspect.getfullargspec(func).kwonlyargs
    )
    if len(arguments) != 2:
        raise ValueError(
            f"Function '{func.__name__}' must have exactly 2 arguments"
        )


def _assert_first_arg_is_dataset(self_arg):
    from climatrix import BaseClimatrixDataset

    if not isinstance(self_arg, BaseClimatrixDataset):
        raise TypeError("Expected first argument to be BaseClimatrixDataset")


def _get_arguments(func, *args, **kwargs) -> tuple:
    sig = inspect.signature(func)
    args_ = sig.bind(*args, **kwargs)
    args_.apply_defaults()
    arg_1, arg_2 = list(args_.arguments.values())
    return arg_1, arg_2


def _validate_args_types(ds1, ds2):
    from climatrix import BaseClimatrixDataset

    if not isinstance(ds1, BaseClimatrixDataset):
        if not isinstance(ds2, BaseClimatrixDataset):
            raise ValueError(
                "At least one argument must be a BaseClimatrixDataset"
            )
    else:
        if not isinstance(ds2, Number):
            raise ValueError(
                "The other argument must be BaseClimatrixDataset or a number"
            )


def _maybe_validate_domains(ds1, ds2):
    from climatrix import BaseClimatrixDataset

    if isinstance(ds2, BaseClimatrixDataset) and ds1.domain != ds2.domain:
        raise ValueError(
            "Domains for both BaseClimatrixDatasets must be the same."
        )


def cm_arithmetic_binary_operator(func):
    """
    Decorator to facilitates arithmetic binary operators for dataset.

    It enables xarray DataArray arithmetic operations between two
    BaseClimatrixDataset instances.
    """
    _assert_is_binary_func(func)

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        from climatrix import BaseClimatrixDataset

        self_arg, other_arg = _get_arguments(func, *args, **kwargs)
        _assert_first_arg_is_dataset(self_arg)
        _maybe_validate_domains(self_arg, other_arg)
        other_arg = other_arg.transpose(*self_arg.dims)
        if isinstance(other_arg, BaseClimatrixDataset):
            other_arg = other_arg.da.data
        res = getattr(self_arg.da.data, func.__name__)(other_arg)
        return type(self_arg)(
            self_arg.domain.to_xarray(res, name=self_arg.da.name)
        )

    return wrapper



from __future__ import annotations

import logging
import os
import warnings
from pathlib import Path
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib.axes import Axes

from climatrix.decorators import raise_if_not_installed

if TYPE_CHECKING:
    from climatrix.dataset.base import BaseClimatrixDataset

sns.set_style("darkgrid")

log = logging.getLogger(__name__)


class Comparison:
    """
    Class for comparing two datasets (dense or sparse).

    For sparse domains, uses nearest neighbor matching with optional
    distance thresholds to find corresponding observations.

    Attributes
    ----------
    predicted_dataset : BaseClimatrixDataset
        The predicted/source dataset.
    true_dataset : BaseClimatrixDataset
        The true/target dataset.
    diff : BaseClimatrixDataset
        The difference between the predicted and true datasets.
    distance_threshold : float, optional
        Maximum distance for point correspondence in sparse domains.

    Parameters
    ----------
    predicted_dataset : BaseClimatrixDataset
        The predicted/source dataset.
    true_dataset : BaseClimatrixDataset
        The true/target dataset.
    map_nan_from_source : bool, optional
        If True, the NaN values from the source dataset will be
        mapped to the target dataset. If False, the NaN values
        from the target dataset will be used. Default is None,
        which means `False` for sparse datasets and `True`
        for dense datasets.
    distance_threshold : float, optional
        For sparse domains, maximum distance threshold for considering
        points as corresponding. If None, closest points are always matched.
        Only used when both datasets have sparse domains.
    """

    def __init__(
        self,
        predicted_dataset: BaseClimatrixDataset,
        true_dataset: BaseClimatrixDataset,
        map_nan_from_source: bool | None = None,
        distance_threshold: float | None = None,
    ):
        from climatrix.dataset.base import BaseClimatrixDataset

        if not isinstance(
            predicted_dataset, BaseClimatrixDataset
        ) or not isinstance(true_dataset, BaseClimatrixDataset):
            raise TypeError(
                "Both datasets must be BaseClimatrixDataset objects"
            )
        self.predicted_dataset = predicted_dataset
        self.true_dataset = true_dataset
        self.distance_threshold = distance_threshold
        self._assert_static()
        if predicted_dataset.domain.is_sparse or true_dataset.domain.is_sparse:
            if (
                predicted_dataset.domain.is_sparse
                and true_dataset.domain.is_sparse
            ):
                self.diff = self._compute_sparse_diff()
            else:
                raise ValueError(
                    "Comparison between sparse and dense domains is not supported. "
                    "Both datasets must be either sparse or dense."
                )
        else:
            if map_nan_from_source is None:
                map_nan_from_source = not predicted_dataset.domain.is_sparse
            if map_nan_from_source:
                try:
                    self.predicted_dataset = self.predicted_dataset.mask_nan(
                        self.true_dataset
                    )
                except ValueError as err:
                    log.error(
                        "Error while masking NaN values from source dataset. "
                        "Set `map_nan_from_source` to False to skip this step."
                    )
                    raise ValueError(
                        "Error while masking NaN values from source dataset. "
                        "Set `map_nan_from_source` to False to skip this step."
                    ) from err
            self.diff = self.predicted_dataset - self.true_dataset

    def _compute_sparse_diff(self) -> BaseClimatrixDataset:
        """
        Compute differences between sparse datasets using nearest neighbor matching.

        Returns
        -------
        BaseClimatrixDataset
            A sparse dataset containing differences between matched points.
        """
        from scipy.spatial import cKDTree

        # Get spatial points for both datasets
        pred_points = self.predicted_dataset.domain.get_all_spatial_points()
        true_points = self.true_dataset.domain.get_all_spatial_points()

        # Build KDTree for true dataset points
        tree = cKDTree(true_points)

        # Find nearest neighbors for predicted dataset points
        if self.distance_threshold is not None:
            if self.distance_threshold == 0.0:
                threshold = np.finfo(float).eps
            else:
                threshold = self.distance_threshold
            distances, indices = tree.query(
                pred_points, distance_upper_bound=threshold
            )
            valid_mask = distances < np.inf
        else:
            distances, indices = tree.query(pred_points)
            valid_mask = np.ones(len(distances), dtype=bool)

        pred_values = self.predicted_dataset.da.values
        true_values = self.true_dataset.da.values

        valid_indices = np.where(valid_mask)[0]
        if len(valid_indices) == 0:
            log.warning(
                "No valid point correspondences found within distance threshold"
            )
            empty_da = self.predicted_dataset.da.isel(
                {self.predicted_dataset.domain.point.name: []}
            )
            return type(self.predicted_dataset)(empty_da)

        result_values = (pred_values - true_values[indices])[valid_indices]

        result_da = self.predicted_dataset.da.isel(
            {self.predicted_dataset.domain.point.name: valid_indices}
        )
        result_da = result_da.copy()
        result_da.values = result_values

        return type(self.predicted_dataset)(result_da)

    def _assert_static(self):
        if (
            self.predicted_dataset.domain.is_dynamic
            or self.true_dataset.domain.is_dynamic
        ):
            raise NotImplementedError(
                "Comparison between dynamic datasets is not yet implemented"
            )

    def plot_diff(
        self,
        title: str | None = None,
        target: str | os.PathLike | Path | None = None,
        show: bool = False,
        ax: Axes | None = None,
        **kwargs,
    ) -> Axes:
        """
        Plot the difference between the source and target datasets.

        Parameters
        ----------
        title : str, optional
            Title of the plot. If not provided, the name of the dataset
            will be used. If the dataset has no name, "Climatrix Dataset" will be used.
        target : str, os.PathLike, Path, or None, optional
            Path to save the plot. If not provided, the plot
            will not be saved.
        show : bool, optional
            Whether to show the plot. Default is False.
        ax : Axes, optional
            Axes to plot on. If not provided, a new figure and axes
            will be created.
        **kwargs : dict, optional
            Additional keyword arguments to pass to the plotting function.

            - `figsize`: tuple, optional
                Size of the figure. Default is (12, 6).
            - `vmin`: float, optional
                Minimum value for the color scale. Default is None.
            - `vmax`: float, optional
                Maximum value for the color scale. Default is None.
            - `cmap`: str, optional
                Colormap to use for the plot. Default is "seismic".
            - `size`: int, optional
                Size of the points for sparse datasets. Default is 10.

        Returns
        -------
        Axes
            The matplotlib axes containing the plot of the difference.
        """
        return self.diff.plot(
            title=title, target=target, show=show, ax=ax, **kwargs
        )

    def plot_signed_diff_hist(
        self,
        ax: Axes | None = None,
        n_bins: int = 50,
        limits: tuple[float] | None = None,
        label: str | None = None,
        alpha: float = 1.0,
    ) -> Axes:
        """
        Plot the histogram of signed difference between datasets.

        The signed difference is a dataset where positive values
        represent areas where the source dataset is larger than
        the target dataset and negative values represent areas
        where the source dataset is smaller than
        the target dataset.

        Parameters
        ----------
        ax : Axes, optional
            The matplotlib axes on which to plot the histogram. If None,
            a new set of axes will be created.
        n_bins : int, optional
            The number of bins to use in the histogram (default is 50).
        limits : tuple[float], optional
            The limits of values to include in the
            histogram (default is None).

        Returns
        -------
        Axes
            The matplotlib axes containing the plot of the signed difference.
        """
        if ax is None:
            fig, ax = plt.subplots()

        ax.hist(
            self.diff.da.values.flatten(),
            bins=n_bins,
            range=limits,
            label=label,
            alpha=alpha,
        )
        return ax

    def compute_rmse(self) -> float:
        """
        Compute the RMSE between the source and target datasets.

        Returns
        -------
        float
            The RMSE between the source and target datasets.
        """
        nanmean = np.nanmean(np.power(self.diff.da.values, 2.0))
        return np.power(nanmean, 0.5).item()

    def compute_mae(self) -> float:
        """
        Compute the MAE between the source and target datasets.

        Returns
        -------
        float
            The mean absolute error between the source and target datasets.
        """
        return np.nanmean(np.abs(self.diff.da.values)).item()

    def compute_mse(self) -> float:
        """
        Compute the MSE between the source and target datasets.

        Returns
        -------
        float
            The mean squared error between the source and target datasets.
        """
        return np.nanmean(np.power(self.diff.da.values, 2.0)).item()

    def compute(self, metric: str) -> float:
        """
        Compute the specified metric.

        Parameters
        ----------
        metric : str
            The metric to compute. Supported values:
            "mae", "mse", "rmse".

        Returns
        -------
        float
            The computed metric value.

        Raises
        ------
        ValueError
            If the metric is not supported.
        """
        metric = metric.lower().strip()
        if metric == "mae":
            return self.compute_mae()
        elif metric == "mse":
            return self.compute_mse()
        elif metric == "rmse":
            return self.compute_rmse()
        else:
            raise ValueError(
                f"Unsupported metric: {metric}. "
                "Supported metrics: mae, mse, rmse"
            )

    @raise_if_not_installed("sklearn")
    def compute_r2(self):
        """
        Compute the R^2 between the source and target datasets.

        Returns
        -------
        float
            The R^2 between the source and target datasets.
        """
        from sklearn.metrics import r2_score

        sd = self.predicted_dataset.da.values.flatten()
        sd = sd[~np.isnan(sd)]
        td = self.true_dataset.da.values.flatten()
        td = td[~np.isnan(td)]
        return r2_score(sd, td)

    def compute_max_abs_error(self) -> float:
        """
        Compute the maximum absolute error between datasets.

        Returns
        -------
        float
            The maximum absolute error between the source and
            target datasets.
        """
        return np.nanmax(np.abs(self.diff.da.values)).item()

    def compute_report(self) -> dict[str, float]:
        return {
            "RMSE": self.compute_rmse(),
            "MAE": self.compute_mae(),
            "Max Abs Error": self.compute_max_abs_error(),
            "R^2": self.compute_r2(),
        }

    def save_report(self, target_dir: str | os.PathLike | Path) -> None:
        """
        Save a report of the comparison between passed datasets.

        This method will create a directory at the specified path
        and save a report of the comparison between the source and
        target datasets in that directory. The report will include
        plots of the difference and signed difference between the
        datasets, as well as a csv file with metrics such
        as the RMSE, MAE, and maximum absolute error.

        Parameters
        ----------
        target_dir : str | os.PathLike | Path
            The path to the directory where the report should be saved.
        """
        target_dir = Path(target_dir)
        if target_dir.exists():
            warnings.warn(
                "The target directory already exists and will be overwritten."
            )
        target_dir.mkdir(parents=True, exist_ok=True)
        metrics = self.compute_report()
        pd.DataFrame(metrics, index=[0]).to_csv(
            target_dir / "metrics.csv", index=False
        )
        self.plot_diff().get_figure().savefig(target_dir / "diff.svg")
        self.plot_signed_diff_hist().get_figure().savefig(
            target_dir / "signed_diff_hist.svg"
        )
        plt.close("all")



from unittest.mock import Mock, patch

import numpy as np

from climatrix.plot.core import Plot


class TestPlotOptimization:
    def setup_method(self):
        self.mock_dataset = Mock()
        self.mock_dataset.domain = Mock()
        self.mock_dataset.domain.is_sparse = False
        self.mock_dataset.domain.has_axis.return_value = False

        self.lats = np.linspace(-90, 90, 100)
        self.lons = np.linspace(-180, 180, 200)
        self.mock_dataset.domain.latitude = Mock()
        self.mock_dataset.domain.latitude.values = self.lats
        self.mock_dataset.domain.longitude = Mock()
        self.mock_dataset.domain.longitude.values = self.lons

        self.values = np.random.rand(100, 200)
        self.mock_dataset.da = Mock()
        self.mock_dataset.da.values = self.values

        self.plot = Plot(self.mock_dataset)

    def test_lod_optimization_with_roi(self):
        lat_min, lat_max = -10, 10
        lon_min, lon_max = -20, 20

        lats_opt, lons_opt, values_opt = self.plot._apply_lod_optimization(
            self.lats,
            self.lons,
            self.values,
            zoom_level=5,
            lat_min=lat_min,
            lat_max=lat_max,
            lon_min=lon_min,
            lon_max=lon_max,
        )

        assert np.all(lats_opt >= lat_min - 5)
        assert np.all(lats_opt <= lat_max + 5)
        assert np.all(lons_opt >= lon_min - 5)
        assert np.all(lons_opt <= lon_max + 5)

    def test_empty_roi(self):
        lat_min, lat_max = 95, 100
        lon_min, lon_max = 185, 190

        lats_opt, lons_opt, values_opt = self.plot._apply_lod_optimization(
            self.lats,
            self.lons,
            self.values,
            zoom_level=5,
            lat_min=lat_min,
            lat_max=lat_max,
            lon_min=lon_min,
            lon_max=lon_max,
        )

        assert len(lats_opt) > 0
        assert len(lons_opt) > 0
        assert values_opt.size > 0

    def test_lod_optimization_no_roi(self):
        lats_opt, lons_opt, values_opt = self.plot._apply_lod_optimization(
            self.lats,
            self.lons,
            self.values,
            zoom_level=3,
            lat_min=None,
            lat_max=None,
            lon_min=None,
            lon_max=None,
        )

        assert len(lats_opt) < len(self.lats)
        assert len(lons_opt) < len(self.lons)
        assert values_opt.shape[0] == len(lats_opt)
        assert values_opt.shape[1] == len(lons_opt)

    def test_zoom_level_step_mapping(self):
        lats_high, lons_high, values_high = self.plot._apply_lod_optimization(
            self.lats,
            self.lons,
            self.values,
            zoom_level=8,
            lat_min=None,
            lat_max=None,
            lon_min=None,
            lon_max=None,
        )

        lats_med, lons_med, values_med = self.plot._apply_lod_optimization(
            self.lats,
            self.lons,
            self.values,
            zoom_level=3,
            lat_min=None,
            lat_max=None,
            lon_min=None,
            lon_max=None,
        )

        lats_low, lons_low, values_low = self.plot._apply_lod_optimization(
            self.lats,
            self.lons,
            self.values,
            zoom_level=0,
            lat_min=None,
            lat_max=None,
            lon_min=None,
            lon_max=None,
        )

        assert len(lats_high) >= len(lats_med) >= len(lats_low)
        assert len(lons_high) >= len(lons_med) >= len(lons_low)

    @patch("climatrix.plot.core.LOD_THRESHOLD", new=10)
    def test_1d_data_optimization(self):
        values_1d = np.random.rand(100)

        lats_opt, lons_opt, values_opt = self.plot._apply_lod_optimization(
            self.lats,
            self.lons,
            values_1d,
            zoom_level=2,
            lat_min=None,
            lat_max=None,
            lon_min=None,
            lon_max=None,
        )

        assert len(lats_opt) < len(self.lats)
        assert len(lons_opt) < len(self.lons)
        assert values_opt.size < len(values_1d)

    def test_small_dataset_below_threshold(self):
        small_lats = np.linspace(-90, 90, 10)
        small_lons = np.linspace(-180, 180, 10)
        small_values = np.random.rand(10, 10)

        lats_opt, lons_opt, values_opt = self.plot._apply_lod_optimization(
            small_lats,
            small_lons,
            small_values,
            zoom_level=1,
            lat_min=None,
            lat_max=None,
            lon_min=None,
            lon_max=None,
        )

        np.testing.assert_array_equal(lats_opt, small_lats)
        np.testing.assert_array_equal(lons_opt, small_lons)
        np.testing.assert_array_equal(values_opt, small_values)

    def test_roi_boundary_conditions(self):
        lat_min, lat_max = -90, -80
        lon_min, lon_max = -180, -170

        lats_opt, lons_opt, values_opt = self.plot._apply_lod_optimization(
            self.lats,
            self.lons,
            self.values,
            zoom_level=5,
            lat_min=lat_min,
            lat_max=lat_max,
            lon_min=lon_min,
            lon_max=lon_max,
        )

        assert len(lats_opt) > 0
        assert len(lons_opt) > 0
        assert values_opt.size > 0

    def test_roi_exact_match(self):
        lat_min = self.lats[10]
        lat_max = self.lats[20]
        lon_min = self.lons[30]
        lon_max = self.lons[40]

        lats_opt, lons_opt, values_opt = self.plot._apply_lod_optimization(
            self.lats,
            self.lons,
            self.values,
            zoom_level=8,
            lat_min=lat_min,
            lat_max=lat_max,
            lon_min=lon_min,
            lon_max=lon_max,
        )

        assert np.min(lats_opt) <= lat_min
        assert np.max(lats_opt) >= lat_max
        assert np.min(lons_opt) <= lon_min
        assert np.max(lons_opt) >= lon_max

    def test_roi_single_point(self):
        lat_center = self.lats[50]
        lon_center = self.lons[100]
        tolerance = 0.1

        lats_opt, lons_opt, values_opt = self.plot._apply_lod_optimization(
            self.lats,
            self.lons,
            self.values,
            zoom_level=8,
            lat_min=lat_center - tolerance,
            lat_max=lat_center + tolerance,
            lon_min=lon_center - tolerance,
            lon_max=lon_center + tolerance,
        )

        assert len(lats_opt) > 0
        assert len(lons_opt) > 0

    def test_roi_with_1d_data(self):
        values_1d = np.random.rand(100)

        lat_min, lat_max = -10, 10
        lon_min, lon_max = -20, 20

        lats_opt, lons_opt, values_opt = self.plot._apply_lod_optimization(
            self.lats,
            self.lons,
            values_1d,
            zoom_level=5,
            lat_min=lat_min,
            lat_max=lat_max,
            lon_min=lon_min,
            lon_max=lon_max,
        )

        np.testing.assert_array_equal(lats_opt, self.lats)
        np.testing.assert_array_equal(lons_opt, self.lons)
        np.testing.assert_array_equal(values_opt, values_1d)

    def test_zero_zoom_level(self):
        lats_opt, lons_opt, values_opt = self.plot._apply_lod_optimization(
            self.lats,
            self.lons,
            self.values,
            zoom_level=0,
            lat_min=None,
            lat_max=None,
            lon_min=None,
            lon_max=None,
        )

        expected_lat_count = len(self.lats[::16])
        expected_lon_count = len(self.lons[::16])

        assert len(lats_opt) == expected_lat_count
        assert len(lons_opt) == expected_lon_count

    def test_negative_zoom_level(self):
        lats_opt, lons_opt, values_opt = self.plot._apply_lod_optimization(
            self.lats,
            self.lons,
            self.values,
            zoom_level=-1,
            lat_min=None,
            lat_max=None,
            lon_min=None,
            lon_max=None,
        )

        expected_lat_count = len(self.lats[::16])
        expected_lon_count = len(self.lons[::16])

        assert len(lats_opt) == expected_lat_count
        assert len(lons_opt) == expected_lon_count

    def test_very_high_zoom_level(self):
        lats_opt, lons_opt, values_opt = self.plot._apply_lod_optimization(
            self.lats,
            self.lons,
            self.values,
            zoom_level=20,
            lat_min=None,
            lat_max=None,
            lon_min=None,
            lon_max=None,
        )

        assert len(lats_opt) == len(self.lats)
        assert len(lons_opt) == len(self.lons)

    def test_partial_roi_parameters(self):
        lats_opt, lons_opt, values_opt = self.plot._apply_lod_optimization(
            self.lats,
            self.lons,
            self.values,
            zoom_level=5,
            lat_min=-10,
            lat_max=10,
            lon_min=None,
            lon_max=None,
        )

        assert len(lats_opt) > 0
        assert len(lons_opt) > 0

    def test_inverted_roi_bounds(self):
        lats_opt, lons_opt, values_opt = self.plot._apply_lod_optimization(
            self.lats,
            self.lons,
            self.values,
            zoom_level=5,
            lat_min=10,
            lat_max=-10,
            lon_min=20,
            lon_max=-20,
        )

        assert len(lats_opt) > 0
        assert len(lons_opt) > 0

    def test_roi_completely_outside_bounds(self):
        lat_min, lat_max = 100, 200
        lon_min, lon_max = 200, 300

        lats_opt, lons_opt, values_opt = self.plot._apply_lod_optimization(
            self.lats,
            self.lons,
            self.values,
            zoom_level=5,
            lat_min=lat_min,
            lat_max=lat_max,
            lon_min=lon_min,
            lon_max=lon_max,
        )

        assert len(lats_opt) > 0
        assert len(lons_opt) > 0

    def test_empty_data_arrays(self):
        empty_lats = np.array([])
        empty_lons = np.array([])
        empty_values = np.array([]).reshape(0, 0)

        lats_opt, lons_opt, values_opt = self.plot._apply_lod_optimization(
            empty_lats,
            empty_lons,
            empty_values,
            zoom_level=5,
            lat_min=None,
            lat_max=None,
            lon_min=None,
            lon_max=None,
        )

        assert len(lats_opt) == 0
        assert len(lons_opt) == 0
        assert values_opt.size == 0

    def test_single_point_data(self):
        single_lat = np.array([0.0])
        single_lon = np.array([0.0])
        single_value = np.array([[1.0]])

        lats_opt, lons_opt, values_opt = self.plot._apply_lod_optimization(
            single_lat,
            single_lon,
            single_value,
            zoom_level=5,
            lat_min=None,
            lat_max=None,
            lon_min=None,
            lon_max=None,
        )

        # Should return the single point (below threshold)
        assert len(lats_opt) == 1
        assert len(lons_opt) == 1
        assert values_opt.size == 1

    def test_nan_values_handling(self):
        """Test handling of NaN values in data"""
        # Create data with NaN values
        values_with_nan = self.values.copy()
        values_with_nan[10:20, 30:40] = np.nan

        lats_opt, lons_opt, values_opt = self.plot._apply_lod_optimization(
            self.lats,
            self.lons,
            values_with_nan,
            zoom_level=5,
            lat_min=None,
            lat_max=None,
            lon_min=None,
            lon_max=None,
        )

        # Should handle NaN values gracefully
        assert len(lats_opt) > 0
        assert len(lons_opt) > 0
        assert values_opt.size > 0

    def test_infinite_values_handling(self):
        """Test handling of infinite values in data"""
        # Create data with infinite values
        values_with_inf = self.values.copy()
        values_with_inf[10:20, 30:40] = np.inf
        values_with_inf[50:60, 70:80] = -np.inf

        lats_opt, lons_opt, values_opt = self.plot._apply_lod_optimization(
            self.lats,
            self.lons,
            values_with_inf,
            zoom_level=5,
            lat_min=None,
            lat_max=None,
            lon_min=None,
            lon_max=None,
        )

        # Should handle infinite values gracefully
        assert len(lats_opt) > 0
        assert len(lons_opt) > 0
        assert values_opt.size > 0

    def test_large_dataset_performance(self):
        """Test optimization with very large dataset"""
        # Create a large dataset
        large_lats = np.linspace(-90, 90, 500)
        large_lons = np.linspace(-180, 180, 1000)
        large_values = np.random.rand(500, 1000)  # 500k points

        lats_opt, lons_opt, values_opt = self.plot._apply_lod_optimization(
            large_lats,
            large_lons,
            large_values,
            zoom_level=1,
            lat_min=None,
            lat_max=None,
            lon_min=None,
            lon_max=None,
        )

        # Should significantly reduce the data size
        assert len(lats_opt) < len(large_lats)
        assert len(lons_opt) < len(large_lons)
        assert values_opt.size < large_values.size

    def test_data_type_preservation(self):
        """Test that data types are preserved after optimization"""
        # Test with different data types
        float32_values = self.values.astype(np.float32)

        lats_opt, lons_opt, values_opt = self.plot._apply_lod_optimization(
            self.lats,
            self.lons,
            float32_values,
            zoom_level=5,
            lat_min=None,
            lat_max=None,
            lon_min=None,
            lon_max=None,
        )

        # Data should maintain its structure (though type might change due to slicing)
        assert isinstance(lats_opt, np.ndarray)
        assert isinstance(lons_opt, np.ndarray)
        assert isinstance(values_opt, np.ndarray)

    def test_step_size_boundaries(self):
        """Test all step size boundaries in zoom level mapping"""
        test_cases = [
            (10, 1),  # zoom_level >= 8
            (8, 1),  # zoom_level >= 8 (boundary)
            (7, 2),  # zoom_level >= 5
            (5, 2),  # zoom_level >= 5 (boundary)
            (4, 4),  # zoom_level >= 3
            (3, 4),  # zoom_level >= 3 (boundary)
            (2, 8),  # zoom_level >= 1
            (1, 8),  # zoom_level >= 1 (boundary)
            (0, 16),  # else case
        ]

        for zoom_level, expected_step in test_cases:
            lats_opt, lons_opt, values_opt = self.plot._apply_lod_optimization(
                self.lats,
                self.lons,
                self.values,
                zoom_level=zoom_level,
                lat_min=None,
                lat_max=None,
                lon_min=None,
                lon_max=None,
            )

            expected_lat_count = len(self.lats[::expected_step])
            expected_lon_count = len(self.lons[::expected_step])

            assert (
                len(lats_opt) == expected_lat_count
            ), f"Zoom {zoom_level}, step {expected_step}"
            assert (
                len(lons_opt) == expected_lon_count
            ), f"Zoom {zoom_level}, step {expected_step}"



import numpy as np
import pyproj
import rioxarray
from scipy import ndimage

# 1. Load DSM raster (assumed to be in EPSG:3035)
dsm_path = "/home/jakub/Downloads/copernicus_DSM_1000m_mood_bbox_epsg3035.tif"
dsm = rioxarray.open_rasterio(
    dsm_path, masked=True
).squeeze()  # remove band dim if any

# 2. Calculate slope (degrees) from DSM elevation data
# Using simple finite differences (gradient) with spacing from CRS units (meters)
# slope = arctan(sqrt(dz/dx^2 + dz/dy^2)) in degrees

# Get pixel size in meters (assume square pixels)
res_x = abs(dsm.rio.resolution()[0])
res_y = abs(dsm.rio.resolution()[1])

dsm.values = dsm.values / 1_000

# Calculate gradients in x and y direction
dz_dx = ndimage.sobel(dsm.values, axis=1) / (8 * res_x)  # axis=1 is x
dz_dy = ndimage.sobel(dsm.values, axis=0) / (8 * res_y)  # axis=0 is y

# Calculate slope in radians, then convert to degrees
slope_rad = np.arctan(np.sqrt(dz_dx**2 + dz_dy**2))
slope_deg = np.degrees(slope_rad)

# Mask slope where DSM is masked or NaN
valid_mask = ~np.isnan(dsm.values) & (dsm.values != dsm.rio.nodata)

# 3. Prepare coordinate transformer (EPSG:3035 -> EPSG:4326)
transformer = pyproj.Transformer.from_crs(
    dsm.rio.crs, "EPSG:4326", always_xy=True
)

# Extract x/y coordinates (meters)
x = dsm.coords["x"].values
y = dsm.coords["y"].values

# Create meshgrid of pixel centers
xx, yy = np.meshgrid(x, y)

# Flatten arrays
xx_flat = xx.flatten()
yy_flat = yy.flatten()
slope_flat = slope_deg.flatten()
elevation_flat = dsm.values.flatten()
valid_flat = valid_mask.flatten()

# Filter valid points
xx_valid = xx_flat[valid_flat]
yy_valid = yy_flat[valid_flat]
slope_valid = slope_flat[valid_flat]
elevation_valid = elevation_flat[valid_flat]

# 4. Transform valid coordinates to lat/lon degrees
lon_valid, lat_valid = transformer.transform(xx_valid, yy_valid)

# 5. Stack slope_results into Nx3 array: [lat, lon, slope]
slope_result = np.vstack([lat_valid, lon_valid, slope_valid]).T
elevation_result = np.vstack([lat_valid, lon_valid, elevation_valid]).T

print(f"Valid points count: {slope_result.shape[0]}")
print("Sample points (lat, lon, slope):")

# 6. Save slope_result as numpy array
mask = np.all(
    [np.isfinite(slope_result[:, -1]), slope_result[:, -1] >= 0], axis=0
)
slope_result = slope_result[mask]
elevation_result = elevation_result[mask]

np.save("lat_lon_elevation.npy", elevation_result)
np.save("lat_lon_slope.npy", slope_result)



import importlib
import importlib.resources
from logging import config

from ._version import __version__


def _configure_logger():
    logging_conf_path = importlib.resources.files("climatrix").joinpath(
        "resources", "logging.ini"
    )
    config.fileConfig(logging_conf_path)


_configure_logger()

from . import optim as optim
from . import plot as plot
from .comparison import Comparison as Comparison
from .dataset.axis import Axis as Axis
from .dataset.axis import AxisType as AxisType
from .dataset.base import BaseClimatrixDataset as BaseClimatrixDataset
from .dataset.domain import Domain as Domain


def seed_all(seed: int):
    """
    Set the seed for all libraries used in the project.
    """
    import random

    import numpy as np

    try:
        import torch
    except ImportError:
        pass
    else:
        torch.manual_seed(seed)
        if torch.cuda.is_available():
            torch.cuda.manual_seed_all(seed)

    random.seed(seed)
    np.random.seed(seed)



from collections import namedtuple

import torch
import torch.nn.functional as F

LossEntity = namedtuple("LossEntity", ["mse", "eikonal", "laplace"])


def compute_sdf_losses(
    xy: torch.Tensor, pred_z: torch.Tensor, true_z: torch.Tensor
) -> LossEntity:
    pred_z = pred_z.squeeze()
    true_z = true_z.squeeze()
    mse_loss = F.mse_loss(pred_z, true_z)

    grad_outputs = torch.ones_like(pred_z)
    grad = torch.autograd.grad(
        pred_z, [xy], grad_outputs=grad_outputs, create_graph=True
    )[0]
    eikonal_loss = torch.abs(grad.norm(dim=-1) - 1).mean()

    grad2 = torch.autograd.grad(grad.sum(), [xy], create_graph=True)[0]
    laplace_loss = (grad2**2).norm(dim=-1).mean()

    return LossEntity(mse=mse_loss, eikonal=eikonal_loss, laplace=laplace_loss)



import importlib
import importlib.resources
import logging
import os
import threading
import time
import webbrowser

import numpy as np
from flask import (
    Flask,
    jsonify,
    render_template_string,
    request,
    send_from_directory,
    url_for,
)

from climatrix.dataset.base import BaseClimatrixDataset
from climatrix.dataset.domain import AxisType

LOD_THRESHOLD: int = int(os.environ.get("LOD_THRESHOLD", 10_000))


class Plot:
    def __init__(self, dataset: BaseClimatrixDataset):
        self.dataset = dataset.to_signed_longitude()
        self.global_min = self.dataset.da.min().item()
        self.global_max = self.dataset.da.max().item()
        self.app = Flask(__name__)
        self.app.logger.setLevel(logging.WARNING)
        self.setup_routes()
        self.app.static_folder = importlib.resources.files(
            "climatrix.resources"
        ).joinpath("static")

    def setup_routes(self):
        @self.app.route("/")
        def index():
            logo_url = url_for("serve_mylibrary_asset", filename="logo.svg")
            return render_template_string(
                self.get_html_template(), logo_url=logo_url
            )

        @self.app.route("/climatrix_assets/<path:filename>")
        def serve_mylibrary_asset(filename):
            return send_from_directory(self.app.static_folder, filename)

        @self.app.route("/api/data")
        def get_data():
            time_idx = request.args.get("time_idx", 0, type=int)
            vertical_idx = request.args.get("vertical_idx", 0, type=int)
            zoom_level = request.args.get("zoom", 1, type=float)

            lat_min = request.args.get("lat_min", type=float)
            lat_max = request.args.get("lat_max", type=float)
            lon_min = request.args.get("lon_min", type=float)
            lon_max = request.args.get("lon_max", type=float)

            data_load = self.prepare_data(
                time_idx,
                vertical_idx,
                zoom_level,
                lat_min,
                lat_max,
                lon_min,
                lon_max,
            )
            return jsonify(data_load)

        @self.app.route("/api/metadata")
        def get_metadata():
            return jsonify(self.get_metadata())

    def get_metadata(self):
        metadata = {
            "has_time": self.dataset.domain.has_axis(AxisType.TIME),
            "has_vertical": self.dataset.domain.has_axis(AxisType.VERTICAL),
            "is_sparse": self.dataset.domain.is_sparse,
        }

        if metadata["has_time"]:
            metadata["time_values"] = [
                str(t) for t in self.dataset.domain.time.values
            ]
            metadata["time_count"] = len(self.dataset.domain.time.values)

        if metadata["has_vertical"]:
            metadata["vertical_values"] = (
                self.dataset.domain.vertical.values.tolist()
            )
            metadata["vertical_count"] = len(
                self.dataset.domain.vertical.values
            )
            metadata["vertical_name"] = self.dataset.domain.vertical.name

        return metadata

    def prepare_data(
        self,
        time_idx=0,
        vertical_idx=0,
        zoom_level=1,
        lat_min=None,
        lat_max=None,
        lon_min=None,
        lon_max=None,
    ):

        if self.dataset.domain.is_sparse:
            return self.prepare_sparse_data(time_idx, vertical_idx)
        else:
            return self.prepare_dense_data(
                time_idx,
                vertical_idx,
                zoom_level,
                lat_min,
                lat_max,
                lon_min,
                lon_max,
            )

    def prepare_sparse_data(self, time_idx, vertical_idx):
        lats = self.dataset.domain.latitude.values
        lons = self.dataset.domain.longitude.values

        data_slice = self.dataset.da
        if self.dataset.domain.has_axis(AxisType.TIME):
            data_slice = data_slice.isel(
                {self.dataset.domain.time.name: time_idx}
            )
        if self.dataset.domain.has_axis(AxisType.VERTICAL):
            if len(data_slice.shape) > 1:
                data_slice = data_slice.isel(
                    {self.dataset.domain.vertical.name: vertical_idx}
                )

        valid_idx = np.isfinite(data_slice.values)
        lats = lats[valid_idx]
        lons = lons[valid_idx]
        data_slice = data_slice[valid_idx]

        return {
            "type": "scatter",
            "lats": lats.tolist(),
            "lons": lons.tolist(),
            "values": data_slice.values.flatten().tolist(),
            "min_val": self.global_min,
            "max_val": self.global_max,
        }

    def prepare_dense_data(
        self,
        time_idx,
        vertical_idx,
        zoom_level=1,
        lat_min=None,
        lat_max=None,
        lon_min=None,
        lon_max=None,
    ):
        lats = self.dataset.domain.latitude.values
        lons = self.dataset.domain.longitude.values

        data_slice = self.dataset.da
        if self.dataset.domain.has_axis(AxisType.TIME):
            data_slice = data_slice.isel(
                {self.dataset.domain.time.name: time_idx}
            )
        if self.dataset.domain.has_axis(AxisType.VERTICAL):
            if len(data_slice.shape) > 2:
                data_slice = data_slice.isel(
                    {self.dataset.domain.vertical.name: vertical_idx}
                )

        values = data_slice.values

        lats_optimized, lons_optimized, values_optimized = (
            self._apply_lod_optimization(
                lats,
                lons,
                values,
                zoom_level,
                lat_min,
                lat_max,
                lon_min,
                lon_max,
            )
        )
        values_cleaned = np.where(
            np.isnan(values_optimized), None, values_optimized
        ).tolist()

        return {
            "type": "mesh",
            "lats": lats_optimized.tolist(),
            "lons": lons_optimized.tolist(),
            "values": values_cleaned,
            "min_val": self.global_min,
            "max_val": self.global_max,
        }

    def _apply_lod_optimization(
        self,
        lats,
        lons,
        values,
        zoom_level,
        lat_min,
        lat_max,
        lon_min,
        lon_max,
    ):

        if len(values.shape) == 2:
            total_points = values.shape[0] * values.shape[1]
        else:
            total_points = len(values)

        if total_points <= LOD_THRESHOLD:
            return lats, lons, values

        if zoom_level >= 8:
            step = 1
        elif zoom_level >= 5:
            step = 2
        elif zoom_level >= 3:
            step = 4
        elif zoom_level >= 1:
            step = 8
        else:
            step = 16

        if all(
            param is not None for param in [lat_min, lat_max, lon_min, lon_max]
        ):
            lat_mask = (lats >= lat_min) & (lats <= lat_max)
            lon_mask = (lons >= lon_min) & (lons <= lon_max)

            if len(values.shape) == 2:
                lat_indices = np.where(lat_mask)[0]
                lon_indices = np.where(lon_mask)[0]

                if len(lat_indices) == 0 or len(lon_indices) == 0:
                    return (
                        lats[:: step * 4],
                        lons[:: step * 4],
                        values[:: step * 4, :: step * 4],
                    )

                lat_start = lat_indices[0]
                lat_end = lat_indices[-1] + 1
                lon_start = lon_indices[0]
                lon_end = lon_indices[-1] + 1

                lat_roi_indices = slice(lat_start, lat_end, step)
                lon_roi_indices = slice(lon_start, lon_end, step)

                return (
                    lats[lat_roi_indices],
                    lons[lon_roi_indices],
                    values[lat_roi_indices, lon_roi_indices],
                )
            else:
                combined_mask = lat_mask & lon_mask
                masked_indices = np.where(combined_mask)[0]
                if len(masked_indices) == 0:
                    return (
                        lats[:: step * 4],
                        lons[:: step * 4],
                        values[:: step * 4],
                    )

                stepped_indices = masked_indices[::step]
                return (
                    lats[stepped_indices],
                    lons[stepped_indices],
                    values[stepped_indices],
                )

        if len(values.shape) == 2:
            return lats[::step], lons[::step], values[::step, ::step]
        else:
            return lats[::step], lons[::step], values[::step]

    def get_html_template(self):
        return (
            importlib.resources.files("climatrix.resources")
            .joinpath("static", "plot_template.html")
            .read_text(encoding="utf8")
        )

    def show(self, port=5000, debug=False):

        def run_server():
            self.app.run(
                host="localhost", port=port, debug=debug, use_reloader=False
            )

        server_thread = threading.Thread(target=run_server)
        server_thread.daemon = True
        server_thread.start()

        time.sleep(1)
        webbrowser.open(f"http://localhost:{port}")

        print(
            f"Climate data visualization server started at http://localhost:{port}"
        )
        print("Press Ctrl+C to stop the server")

        try:
            while True:
                time.sleep(1)
        except KeyboardInterrupt:
            print("\nShutting down server...")



from __future__ import annotations

import logging
import os
import warnings
from datetime import datetime
from itertools import product
from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterable, Self

import cartopy.crs as ccrs
import cartopy.feature as cfeature
import numpy as np
import xarray as xr
from matplotlib import pyplot as plt
from matplotlib.axes import Axes

from climatrix.dataset.axis import Axis, AxisType
from climatrix.dataset.domain import (
    Domain,
    SamplingNaNPolicy,
    ensure_single_var,
)
from climatrix.dataset.utils import ensure_list_or_slice
from climatrix.decorators import cm_arithmetic_binary_operator
from climatrix.exceptions import (
    DomainMismatchError,
    LongitudeConventionMismatch,
    SubsettingByNonDimensionAxisError,
)
from climatrix.types import Latitude, Longitude

if TYPE_CHECKING:
    from climatrix.reconstruct.type import ReconstructionType


def drop_scalar_coords_and_dims(da: xr.DataArray) -> xr.DataArray:
    coords_and_dims = {*da.dims, *da.coords.keys()}
    for coord in coords_and_dims:
        if len(da[coord].shape) == 0:
            da = da.drop_vars(coord)
    return da


log = logging.getLogger(__name__)


@xr.register_dataset_accessor("cm")
@xr.register_dataarray_accessor("cm")
class BaseClimatrixDataset:
    """
    Base class for Climatrix workflows.

    This class provides a set of methods for manipulating
    xarray datasets. It is designed to be used as an
    xarray accessor, allowing you to call its methods
    directly on xarray datasets.

    The class supports basic arithmetic operations, including:
    addition, subtraction, multiplication, and division.

    Attributes
    ----------
    da : xarray.DataArray
        The underlying `xarray.DataArray` object (if
        single-variable `xarray.Dataset` was passed, it is squeezed
        to `xarray.DataArray`).
    domain : Domain
        The domain object representing the spatial
        and temporal dimensions of the dataset.
        See [`SparseDomain`][climatrix.dataset.domain.SparseDomain] and
        [`DenseDomain`][climatrix.dataset.domain.DenseDomain]
        for more details.
    """

    __slots__ = (
        "da",
        "domain",
    )

    da: xr.DataArray
    domain: Domain

    def __init__(self, xarray_obj: xr.DataArray):
        # NOTE: At the moment we only support DataArray or
        # Dataset with a single variable
        xarray_obj = ensure_single_var(xarray_obj)
        xarray_obj = drop_scalar_coords_and_dims(xarray_obj)
        self.domain = Domain(xarray_obj)
        self.da = xarray_obj

    # ###############################
    # Properties
    # ###############################
    @property
    def dims(self) -> tuple[AxisType, ...]:
        """
        Get the dimensions of the dataset.

        Returns
        -------
        tuple[AxisType, ...]
            A tuple of `AxisType` objects representing the dimensions
            of the dataset.
        """
        return self.domain.dims

    # ###############################
    #  Operators
    # ###############################
    @cm_arithmetic_binary_operator
    def __add__(self, other: Any) -> Self: ...  # noqa: E704

    @cm_arithmetic_binary_operator
    def __sub__(self, other: Any) -> Self: ...  # noqa: E704

    @cm_arithmetic_binary_operator
    def __mul__(self, other: Any) -> Self: ...  # noqa: E704

    @cm_arithmetic_binary_operator
    def __truediv__(self, other: Any) -> Self: ...  # noqa: E704

    # ###############################
    #  Rolling
    # ###############################
    def to_signed_longitude(self) -> Self:
        """
        Convert the dataset to signed longitude convention.

        The longitude values are converted to be in
        the range (-180 to 180 degrees).

        Examples
        --------
        >>> import climatrix as cm
        >>> dset = xr.open_dataset("path/to/dataset.nc").cm
        >>> dset.da
        <xarray.DataArray 'temperature' (time: 1, latitude: 180, longitude: 360)>
        ...
        Dimensions:  (time: 1, latitude: 180, longitude: 360)
        Coordinates:
          * time     (time) datetime64[ns] 2020-01-01
          * latitude (latitude) float64 -90.0 -89.0 -88.0 ... 88.0 89.0
          * longitude (longitude) float64 0.0 1.0 2.0 ... 357.0 358.0 359.0
        Data variables:
            temperature (time, latitude, longitude) float64 ...
        >>> dset2 = cm.to_signed_longitude()
        >>> dset2.da
        <xarray.DataArray 'temperature' (time: 1, latitude: 180, longitude: 360)>
        ...
        Dimensions:  (time: 1, latitude: 180, longitude: 360)
        Coordinates:
          * time     (time) datetime64[ns] 2020-01-01
          * latitude (latitude) float64 -90.0 -89.0 -88.0 ... 88.0 89.0
          * longitude (longitude) float64 -180.0 -179.0 -178.0 ... 177.0 178.0 179.0

        References
        ----------
        [1] Mancini, M., Walczak, J. Stojiljkovic, M., geokube: A Python
           library for geospatial data processing, 2024,
           [https://doi.org/10.5281/zenodo.10597965](https://doi.org/10.5281/zenodo.10597965)
           [https://github.com/CMCC-Foundation/geokube](https://github.com/CMCC-Foundation/geokube)
        """
        if (
            not self.domain.has_axis(AxisType.LONGITUDE)
            or not self.domain.longitude.is_dimension
        ):
            return type(self)(self.da)
        roll_value = (self.da[self.domain.longitude.name] >= 180).sum().item()
        res = self.da.assign_coords(
            {
                self.domain.longitude.name: (
                    ((self.da[self.domain.longitude.name] + 180) % 360) - 180
                )
            }
        ).roll(**{self.domain.longitude.name: roll_value}, roll_coords=True)
        res[self.domain.longitude.name].attrs.update(
            self.da[self.domain.longitude.name].attrs
        )
        return type(self)(res)

    def to_positive_longitude(self) -> Self:
        """
        Convert the dataset to positive longitude convention.

        The longitude values are converted to be in
        the range (0 to 360 degrees).

        Examples
        --------
        >>> import climatrix as cm
        >>> dset = xr.open_dataset("path/to/dataset.nc").cm
        >>> dset.da
        <xarray.DataArray 'temperature' (time: 1, latitude: 180,
        longitude: 360)>
        ...
        Dimensions:  (time: 1, latitude: 180, longitude: 360)
        Coordinates:
          * time     (time) datetime64[ns] 2020-01-01
          * latitude (latitude) float64 -90.0 -89.0 -88.0 ... 88.0 89.0
          * longitude (longitude) float64 -180.0 ... 178.0 179.0
        Data variables:
            temperature (time, latitude, longitude) float64 ...
        >>> dset2 = dset.to_positive_longitude()
        >>> dset2.da
        <xarray.DataArray 'temperature' (time: 1, latitude: 180,
          longitude: 360)>
        ...
        Dimensions:  (time: 1, latitude: 180, longitude: 360)
        Coordinates:
          * time     (time) datetime64[ns] 2020-01-01
          * latitude (latitude) float64 -90.0 -89.0 -88.0 ... 88.0 89.0
          * longitude (longitude) float64 0.0 1.0 ... 357.0 358.0 359.0
        Data variables:
            temperature (time, latitude, longitude) float64 ...

        References
        ----------
        [1] Mancini, M., Walczak, J. Stojiljkovic, M., geokube: A Python
           library for geospatial data processing, 2024,
           [https://doi.org/10.5281/zenodo.10597965](https://doi.org/10.5281/zenodo.10597965)
           [https://github.com/CMCC-Foundation/geokube](https://github.com/CMCC-Foundation/geokube)
        """
        if (
            not self.domain.has_axis(AxisType.LONGITUDE)
            or not self.domain.longitude.is_dimension
        ):
            return type(self)(self.da)
        roll_value = (self.da[self.domain.longitude.name] <= 0).sum().item()
        res = (
            self.da.assign_coords(
                {
                    self.domain.longitude.name: (
                        self.da[self.domain.longitude.name] % 360
                    )
                }
            )
            .roll(
                **{self.domain.longitude.name: -roll_value}, roll_coords=True
            )
            .assign_attrs(**self.da[self.domain.longitude.name].attrs)
        )
        res[self.domain.longitude.name].attrs.update(
            self.da[self.domain.longitude.name].attrs
        )
        return type(self)(res)

    # ###############################
    # Utility methods
    # ###############################

    def squeeze(self) -> Self:
        """
        Squeeze the dataset to remove dimensions of size 1.

        Returns
        -------
        Self
            The squeezed dataset.
        """
        da = self.da.squeeze()
        return type(self)(da)

    def profile_along_axes(
        self, *axes: AxisType | str
    ) -> Iterable[BaseClimatrixDataset]:
        """
        Generate profiles along the specified axes.

        Parameters
        ----------
        *axes : AxisType | str
            The axes along which to generate profiles.

        Yields
        ------
        BaseClimatrixDataset
            A dataset containing the profile along the specified axes.
        """
        axes = [AxisType.get(axis) for axis in axes]
        axis_values = []
        for axis in axes:
            axis_obj = self.domain.get_axis(axis)
            if axis_obj is None or axis_obj.values is None:
                continue
            axis_values.append(range(len(axis_obj.values)))

        for combination in product(*axis_values):
            profile = self.isel(
                {axis: value for axis, value in zip(axes, combination)}
            )
            yield profile

    def mask_nan(self, source: Self) -> Self:
        """
        Apply NaN values from another dataset to the current one.

        Parameters
        ----------
        source : BaseClimatrixDataset
            Dataset whose NaN values will be applied to the current one.

        Returns
        -------
        BaseClimatrixDataset
            A new dataset with NaN values applied.

        Raises
        ------
        TypeError
            If the `source` argument is not a BaseClimatrixDataset.
        ValueError
            If the domain of the `source` or the current dataset is sparse.
        DomainMismatchError
            If the domains of the `source` and the current dataset
            do not match.

        Examples
        --------
        >>> import climatrix as cm
        >>> dset1 = xr.open_dataset("path/to/dataset1.nc").cm
        >>> dset2 = xr.open_dataset("path/to/dataset2.nc").cm
        >>> dset1.mask_nan(dset2)
        """
        if not isinstance(source, BaseClimatrixDataset):
            raise TypeError("Argument `source` must be a BaseClimatrixDataset")
        if source.domain.is_sparse or self.domain.is_sparse:
            raise ValueError(
                "Masking NaN values is only supported for dense domain."
            )
        if self.domain != source.domain:
            log.error("Domains of the datasets do not match.")
            raise DomainMismatchError("Domains of the datasets do not match.")

        source = source.transpose(*self.dims)
        # NOTE: there can be slight differences (E.g. due to overflow
        # or floating point precision) in the coordinates of the source,
        source_da = source.da.reindex_like(
            self.da, method="nearest", tolerance=1e-5
        )
        da = xr.where(source_da.isnull(), np.nan, self.da)
        return type(self)(da)

    def transpose(self, *axes: AxisType | str) -> Self:
        """
        Transpose the dataset along the specified dimensions.

        Parameters
        ----------
        *axes : AxisType or str
            The axes along which to transpose the dataset.

        Returns
        -------
        Self
            The transposed dataset.

        Examples
        --------
        >>> import climatrix as cm
        >>> dset = xr.open_dataset("path/to/dataset.nc").cm
        >>> dset2 = dset.transpose("longitude", "latitude")
        """
        dim_names = [self.domain.get_axis(ax).name for ax in axes]
        da = self.da.transpose(*dim_names)
        da = da.compute() if hasattr(da.data, "dask") else da
        return type(self)(da)

    def sel(self, query: dict[AxisType | str, Any]) -> Self:
        """
        Select data along the specified axes.

        Parameters
        ----------
        query : dict
            Dictionary of axes and values to select.
            The keys can be either `AxisType` objects or strings
            associated to standard names of axis type.

        Returns
        -------
        Self
            The selected dataset.

        Examples
        --------
        >>> import climatrix as cm
        >>> from climatrix import AxisType
        >>> dset = xr.open_dataset("path/to/dataset.nc").cm
        >>> dset2 = dset.sel({"latitude": 10.0, "longitude": 20.0})
        >>> dset3 = dset.sel({AxisType.TIME: [datetime(2020, 1, 1)]})
        """
        query_ = {}
        # Handle sparse domain coordinate selection specially
        sparse_coord_queries = {}

        for axis_type, value in query.items():
            axis_type = AxisType.get(axis_type)
            corresponding_axis = self.domain.get_axis(axis_type)
            if corresponding_axis is None:
                log.warning(
                    "Axis for '%s' not found in the dataset", axis_type
                )
                warnings.warn(
                    f"Axis for '{axis_type}' not found in the dataset."
                    "Skipping."
                )
                continue
            if not corresponding_axis.is_dimension:
                if (
                    self.domain.is_sparse
                    and corresponding_axis.name in self.da.coords
                ):
                    coord_dims = self.da.coords[corresponding_axis.name].dims
                    if any(
                        dim_name == self.domain.point.name
                        for dim_name in coord_dims
                    ):
                        sparse_coord_queries[corresponding_axis.name] = value
                        continue
                    else:
                        log.error(
                            "Axis '%s' is not a dimension and does not depend on sparse dimensions in this dataset. ",
                            axis_type,
                        )
                        raise SubsettingByNonDimensionAxisError(
                            f"Axis '{axis_type}' is not a dimension and does not depend on sparse dimensions in this dataset. "
                            "Subsetting by non-dimension axis is only supported for coordinates that depend on sparse dimensions."
                        )
                else:
                    log.error(
                        "Axis '%s' is not a dimension in this dataset. ",
                        axis_type,
                    )
                    raise SubsettingByNonDimensionAxisError(
                        f"Axis '{axis_type}' is not a dimension in this dataset. "
                        "Subsetting by non-dimension axis is not supported."
                    )
            query_[corresponding_axis.name] = ensure_list_or_slice(value)

        # First handle regular dimensional selections
        da = self.da
        if query_:
            da = da.sel(query_)

        # Then handle sparse coordinate selections using where
        for coord_name, coord_value in sparse_coord_queries.items():
            coord_value = ensure_list_or_slice(coord_value)
            if isinstance(coord_value, list):
                # Handle multiple values
                mask = da.coords[coord_name].isin(coord_value)
            else:
                # Handle single value or slice
                if isinstance(coord_value, slice):
                    # Handle slice selection
                    coord_values = da.coords[coord_name]
                    if coord_value.start is not None:
                        mask = coord_values >= coord_value.start
                    else:
                        mask = xr.ones_like(coord_values, dtype=bool)
                    if coord_value.stop is not None:
                        mask = mask & (coord_values <= coord_value.stop)
                else:
                    # Handle exact value
                    mask = da.coords[coord_name] == coord_value
            da = da.where(mask, drop=True)

        return type(self)(da)

    def isel(self, query: dict[Axis | str, Any]) -> Self:
        """
        Select data along the specified axes by index.

        Parameters
        ----------
        query : dict
            Dictionary of axes and values to select.
            The keys can be either `AxisType` objects or strings
            associated to standard names of axis type.
            The values must be integers, list of integers, or slices.

        Returns
        -------
        Self
            The selected dataset.

        Examples
        --------
        >>> import climatrix as cm
        >>> from climatrix import AxisType
        >>> dset = xr.open_dataset("path/to/dataset.nc").cm
        >>> dset2 = dset.isel({"latitude": 0, "longitude": 1})
        >>> dset3 = dset.isel({AxisType.TIME: slice(0, 2)})
        """
        query_ = {}
        for axis_type, value in query.items():
            axis_type = AxisType.get(axis_type)
            corresponding_axis = self.domain.get_axis(axis_type)
            if corresponding_axis is None:
                log.warning(
                    "Axis for '%s' not found in the dataset", axis_type
                )
                warnings.warn(
                    f"Axis for '{axis_type}' not found in the dataset."
                    "Skipping."
                )
                continue
            if not corresponding_axis.is_dimension:
                if (
                    self.domain.is_sparse
                    and corresponding_axis.name in self.da.coords
                ):
                    coord_dims = self.da.coords[corresponding_axis.name].dims
                    if any(
                        dim_name == self.domain.point.name
                        for dim_name in coord_dims
                    ):
                        log.warning(
                            "Using isel with coordinate '%s' in sparse domain. "
                            "Converting to point dimension selection.",
                            axis_type,
                        )
                        warnings.warn(
                            f"Using isel with coordinate '{axis_type}' in sparse domain. "
                            "This selects the point at the given index, not the coordinate value. "
                            "Consider using sel() instead for coordinate value selection."
                        )
                        query_[self.domain.point.name] = ensure_list_or_slice(
                            value
                        )
                        continue
                    else:
                        log.error(
                            "Axis '%s' is not a dimension and does not depend on sparse dimensions in this dataset. ",
                            axis_type,
                        )
                        raise SubsettingByNonDimensionAxisError(
                            f"Axis '{axis_type}' is not a dimension and does not depend on sparse dimensions in this dataset. "
                            "Subsetting by non-dimension axis is only supported for coordinates that depend on sparse dimensions."
                        )
                else:
                    log.error(
                        "Axis '%s' is not a dimension in this dataset. ",
                        axis_type,
                    )
                    raise SubsettingByNonDimensionAxisError(
                        f"Axis '{axis_type}' is not a dimension in this dataset. "
                        "Subsetting by non-dimension axis is not supported."
                    )
            query_[corresponding_axis.name] = ensure_list_or_slice(value)
        da = self.da.isel(query_)
        return type(self)(da)

    # ###############################
    #  Subsetting
    # ###############################
    def subset(
        self,
        north: float | None = None,
        south: float | None = None,
        west: float | None = None,
        east: float | None = None,
    ) -> Self:
        r"""
        Subset data with the specified bounding box.

        If an argument is not provided, it means no bounds set
        in that direction. For example, if `north` is not provided,
        it means that the maximum latitude of the dataset will be used.
        If `north` and `south` are provided, the dataset will be
        subsetted to the area between these two latitudes.

        Parameters
        ----------
        north : float, optional
            North latitude of the bounding box.
        south : float, optional
            South latitude of the bounding box.
        west : float, optional
            West longitude of the bounding box.
        east : float, optional
            East longitude of the bounding box.

        Returns
        -------
        Self
            The subsetted dataset.

        Raises
        ------
        LongitudeConventionMismatch
            - If the dataset is in positive-only convention (longitude $\lambda \in [0, 360]$) and
            negative values are requested, or vice versa.
            - If the dataset is in signed-longitude convention (longitude $\lambda \in [-180, 180]$) and
            positive values greater than 360 are requested.

        Examples
        --------
        >>> import climatrix as cm
        >>> globe_dset = xr.open_dataset("path/to/dataset.nc")
        >>> globe_dset
        <xarray.Dataset>
        Dimensions:  (time: 1, latitude: 180, longitude: 360)
        Coordinates:
          * time     (time) datetime64[ns] 2020-01-01
          * latitude (latitude) float64 -90.0 -89.0 -88.0 ... 88.0 89.0
          * longitude (longitude) float64 0.0 1.0 2.0 ... 357.0 358.0 359.0
        Data variables:
            temperature (time, latitude, longitude) float64 ...
        >>> dset2 = globe_dset.cm.subset(
        ...     north=10.0,
        ...     south=5.0,
        ...     west=20.0,
        ...     east=25.0,
        ... )
        >>> dset2 = globe_dset.cm.subset(
        ...     north=10.0,
        ...     south=5.0,
        ...     west=-50.0,
        ...     east=25.0,
        ... )
        LongitudeConventionMismatch: The dataset is in positive-only convention
        (longitude goes from 0 to 360) while you are
        requesting negative values (longitude goes from -180 to 180).
        """
        idx, start, stop = self.domain._compute_subset_indexers(
            north=north,
            south=south,
            west=west,
            east=east,
        )
        first_el, _ = (
            self.domain.longitude.values.min(),
            self.domain.longitude.values.max(),
        )
        start = 0 if start is None else start
        stop = 0 if stop is None else stop
        sel_neg_conv = (start < 0) | (stop < 0)
        sel_pos_conv = (start > 180) | (stop > 180)

        dset_neg_conv = first_el < 0
        dset_pos_conv = first_el >= 0
        if dset_pos_conv and sel_neg_conv:
            raise LongitudeConventionMismatch(
                "The dataset is in positive-only convention "
                "(longitude goes from 0 to 360) while you are "
                "requesting negative values (longitude goes "
                "from -180 to 180). Run `to_signed_longitude()` "
                "first to convert dataset properly."
            )
        if dset_neg_conv and sel_pos_conv:
            raise LongitudeConventionMismatch(
                "The dataset is in signed-longitude convention "
                "(longitude goes from -180 to 180) while you are "
                "requesting values from 0 to 360. "
                "Run `to_positive_longitude()` first to convert "
                "dataset properly."
            )
        da = self.da.sel(idx)
        if da.size == 0:
            logging.warning(
                "No data found in the specified bounding box. "
                "Returning empty dataset."
            )
            warnings.warn(
                "No data found in the specified bounding box "
                f"(north={north}, south={south}, west={west}, east={east}). "
                "Returning empty dataset."
            )

        return type(self)(da)

    def time(
        self, time: datetime | np.datetime64 | slice | list | np.ndarray
    ) -> Self:
        """
        Select data at a specific time or times.

        Parameters
        ----------
        time : datetime, np.datetime64, slice, list, or np.ndarray
            Time or times to be selected.

        Returns
        -------
        Self
            The dataset with the selected time or times.

        Examples
        --------
        >>> import climatrix as cm
        >>> dset = xr.open_dataset("path/to/dataset.nc")

        Selecting by `datetime` object:
        >>> dset.cm.time(datetime(2020, 1, 1))

        Selecting by `np.datetime64` object:
        >>> dset.cm.time(np.datetime64("2020-01-01"))

        Selecting by `str` object:
        >>> dset.cm.time(slice("2020-01-01"))

        Selecting by `list` of any of the above:
        >>> dset.cm.time([datetime(2020, 1, 1), np.datetime64("2020-01-02")])

        Selecting by `slice` object:
        >>> dset.cm.time(slice(datetime(2020, 1, 1), datetime(2020, 1, 2)))
        """
        time = ensure_list_or_slice(time)
        return type(self)(
            self.da.sel({self.domain.time.name: time}, method="nearest")
        )

    def itime(self, time: int | list[int] | np.ndarray | slice) -> Self:
        """
        Select time value by index.

        Parameters
        ----------
        time : int, list[int], np.ndarray, or slice
            Time index or indices to be selected.

        Returns
        -------
        Self
            The dataset with the selected time or times.

        Examples
        --------
        >>> import climatrix as cm
        >>> dset = xr.open_dataset("path/to/dataset.nc")

        Selecting by `int` object:
        >>> dset.cm.itime(0)

        Selecting by `list` of `int`s:
        >>> dset.cm.itime([0, 1])

        Selecting by `slice` object:
        >>> dset.cm.itime(slice(0, 2))
        """
        if self.domain.is_dynamic:
            time = ensure_list_or_slice(time)
            return type(self)(self.da.isel({self.domain.time.name: time}))
        return self

    # ##############################
    #  Sampling
    # ###############################
    def sample_uniform(
        self,
        portion: float | None = None,
        number: int | None = None,
        nan: SamplingNaNPolicy | str = "ignore",
    ) -> Self:
        """
        Sample the dataset using a uniform distribution.

        Parameters
        ----------
        portion : float, optional
            Portion of the dataset to be sampled.
        number : int, optional
            Number of points to be sampled.
        nan : SamplingNaNPolicy | str, optional
            Policy for handling NaN values.

        Notes
        -----
        At least one of `portion` or `number` must be provided.
        Cannot be provided both at the same time.

        Warns
        -----
        TooLargeSamplePortionWarning
            If the portion exceeds 1.0 or number of points exceeds
            the number of spatial points in the Domain

        Raises
        ------
        ValueError
            If the dataset contains NaN values and `nan` parameter
             (NaN handling policy) is set to `SamplingNaNPolicy.RAISE`.

        Examples
        --------
        >>> import climatrix as cm
        >>> dset = xr.open_dataset("path/to/dataset.nc")
        >>> sparse_dset = dset.cm.sample_uniform(portion=0.1)
        """
        nan = SamplingNaNPolicy.get(nan)
        if nan in (SamplingNaNPolicy.RAISE, SamplingNaNPolicy.IGNORE):
            idx = self.domain._compute_sample_uniform_indexers(
                portion=portion, number=number
            )
        elif nan == SamplingNaNPolicy.RESAMPLE:
            idx = self.domain._compute_sample_no_nans_indexers(
                self.da, portion=portion, number=number
            )
        da = self.da.isel(idx)
        if nan == SamplingNaNPolicy.RAISE and da.isnull().any():
            raise ValueError("Not all points have data")
        return type(self)(da)

    def sample_normal(
        self,
        portion: float | None = None,
        number: int | None = None,
        center_point: tuple[Longitude, Latitude] = None,
        sigma: float = 10.0,
        nan: SamplingNaNPolicy | str = "ignore",
    ) -> Self:
        """
        Sample the dataset using a normal distribution.

        Parameters
        ----------
        portion : float, optional
            Portion of the dataset to be sampled.
        number : int, optional
            Number of points to be sampled.
        center_point : tuple[Longitude, Latitude], optional
            Center point for the normal distribution.
        sigma : float, optional
            Standard deviation for the normal distribution.
        nan : SamplingNaNPolicy | str, optional
            Policy for handling NaN values.

        Notes
        -----
        At least one of `portion` or `number` must be provided.
        Cannot be provided both at the same time.

        Warns
        -----
        TooLargeSamplePortionWarning
            If the portion exceeds 1.0 or number of points exceeds
            the number of spatial points in the Domain

        Raises
        ------
        ValueError
            If the dataset contains NaN values and `nan` parameter
             (NaN handling policy) is set to `SamplingNaNPolicy.RAISE`.

        Examples
        --------
        >>> import climatrix as cm
        >>> dset = xr.open_dataset("path/to/dataset.nc")
        >>> sparse_dset = dset.cm.sample_normal(
        ...     number=1_000,
        ...     center_point=(10.0, 20.0),
        ...     sigma=5.0,
        ... )
        """
        nan = SamplingNaNPolicy.get(nan)
        idx = self.domain._compute_sample_normal_indexers(
            portion=portion,
            number=number,
            center_point=center_point,
            sigma=sigma,
        )
        da = self.da.isel(idx)
        if nan == SamplingNaNPolicy.RAISE and da.isnull().any():
            raise ValueError("Not all points have data")
        return type(self)(da)

    # ###############################
    #  Reconstruction
    # ###############################
    def reconstruct(
        self,
        target: Domain,
        *,
        method: ReconstructionType | str,
        **recon_kwargs,
    ) -> Self:
        """
        Reconstruct the dataset to a target domain.

        If target domain is sparse, the reconstruction will be sparse
        too. If target domain is dense, the reconstruction will be dense
        too. The reconstruction will be done using the method specified
        in the `method` argument.

        The method can be one of the following:
            Inverse Distance Weightining (`idw`),
            Ordinary Kriging (`ok`).

        Parameters
        ----------
        target : Domain
            The target domain to reconstruct the dataset to.
        method : ReconstructionType | str
            The method to use for reconstruction. Can be one of the
            following: 'idw', 'ok'.
        recon_kwargs : dict
            Additional keyword arguments to pass to the reconstruction
            method.

        See Also
        --------
        - [`climatrix.reconstruct.type.ReconstructionType`](/climatrix/api/#climatrix.reconstruct.type.ReconstructionType)
        - [`climatrix.reconstruct.idw.IDWReconstructor`](/climatrix/api/#climatrix.reconstruct.idw.IDWReconstructor)
        - [`climatrix.reconstruct.kriging.OrdinaryKrigingReconstructor`](/climatrix/api/#climatrix.reconstruct.kriging.OrdinaryKrigingReconstructor)
        - [`climatrix.reconstruct.siren.siren.SIRENReconstructor`](/climatrix/api/#climatrix.reconstruct.siren.siren.SIRENReconstructor)

        Returns
        -------
        Self
            The reconstructed dataset.
        """
        from climatrix.reconstruct.type import ReconstructionType

        method = ReconstructionType.get(method)
        return (
            ReconstructionType.get(method)
            .value(self, target_domain=target, **recon_kwargs)
            .reconstruct()
        )

    # ###############################
    #  Plotting
    # ###############################
    def plot(
        self,
        title: str | None = None,
        target: str | os.PathLike | Path | None = None,
        show: bool = True,
        **kwargs,
    ) -> Axes:
        """
        Plot the dataset on a map.

        The dataset is plotted using Cartopy and Matplotlib.

        Parameters
        ----------
        title : str, optional
            Title of the plot. If not provided, the name of the dataset
            will be used. If the dataset has no name, "Climatrix Dataset" will be used.
        target : str, os.PathLike, Path, or None, optional
            Path to save the plot. If not provided, the plot
            will not be saved.
        show : bool, optional
            Whether to show the plot. Default is True.
        **kwargs : dict, optional
            Additional keyword arguments to pass to the plotting function.

            - `figsize`: tuple, optional
                Size of the figure. Default is (12, 6).
            - `vmin`: float, optional
                Minimum value for the color scale. Default is None.
            - `vmax`: float, optional
                Maximum value for the color scale. Default is None.
            - `cmap`: str, optional
                Colormap to use for the plot. Default is "seismic".
            - `ax`: Axes, optional
                Axes to plot on. If not provided, a new figure and axes
                will be created.
            - `size`: int, optional
                Size of the points for sparse datasets. Default is 10.

        Returns
        -------
        Axes
            The axes object containing the plot.

        Raises
        ------
        NotImplementedError
            If the dataset is dynamic (contains time dimension
            with more than one value).
        """
        for axis in self.domain.all_axes_types:
            if axis in [AxisType.LATITUDE, AxisType.LONGITUDE, AxisType.POINT]:
                continue
            elif (
                self.domain.get_axis(axis).is_dimension
                and self.domain.get_size(axis) > 1
            ):
                warnings.warn(
                    f"Dataset contains axis '{axis}' with more than one "
                    "value. It is not yet supported for plotting. "
                    "The first value will be used."
                )
                self.isel({axis: 0}).plot()
                return
        figsize = kwargs.pop("figsize", (12, 6))
        vmin = kwargs.pop("vmin", None)
        vmax = kwargs.pop("vmax", None)
        cmap = kwargs.pop("cmap", "seismic")
        ax = kwargs.pop("ax", None)
        cbar_name = kwargs.pop("cbar_name", None)
        title = title or self.da.name or "Climatrix Dataset"

        lat = self.da[self.domain.latitude.name]
        lon = self.da[self.domain.longitude.name]
        proj = ccrs.PlateCarree()

        if ax is None:
            fig, ax = plt.subplots(
                figsize=figsize, subplot_kw={"projection": proj}
            )
            ax.coastlines()
            ax.add_feature(cfeature.BORDERS, linewidth=0.5)
            ax.add_feature(cfeature.LAND, facecolor="lightgray")
            ax.add_feature(cfeature.OCEAN, facecolor="lightblue")
            ax.gridlines(draw_labels=True, linewidth=0.2, linestyle="--")

            ax.text(
                -0.07,
                0.55,
                "latitude",
                va="bottom",
                ha="center",
                rotation="vertical",
                rotation_mode="anchor",
                transform=ax.transAxes,
            )
            ax.text(
                0.5,
                -0.1,
                "longitude",
                va="bottom",
                ha="center",
                rotation="horizontal",
                rotation_mode="anchor",
                transform=ax.transAxes,
            )

            ax.set_aspect("equal")

        if self.domain.is_sparse:
            size = kwargs.pop("size", 10)
            actor = ax.scatter(
                lon,
                lat,
                c=self.da.squeeze(),
                cmap=cmap,
                vmin=vmin,
                vmax=vmax,
                transform=proj,
                marker="x",
                s=size,
            )
        else:
            actor = ax.pcolormesh(
                lon,
                lat,
                self.da.squeeze(),
                transform=proj,
                cmap=cmap,
                shading="auto",
                vmin=vmin,
                vmax=vmax,
            )

        cbar = plt.colorbar(
            actor, ax=ax, orientation="vertical", shrink=0.7, pad=0.05
        )
        cbar.set_label(cbar_name or self.da.name or "Value")

        if title:
            ax.set_title(title, fontsize=14)

        plt.tight_layout()
        if target is not None:
            target = Path(target)
            target.parent.mkdir(parents=True, exist_ok=True)
            plt.savefig(target, dpi=300)
        if show:
            plt.show()
        return ax



import numpy as np
import torch
import torch.nn as nn


class SineLayer(nn.Module):
    """
    Sine activation layer with custom frequency and initialization.

    Based on https://github.com/vsitzmann/siren
    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool = True,
        is_first: bool = False,
        omega_0: float = 30,
    ) -> None:
        """
        Initialize sine activation layer with frequency omega_0.

        Parameters
        ----------
        in_features : int
            Number of input features
        out_features : int
            Number of output features
        bias : bool, default=True
            Whether to use bias in the linear layer
        is_first : bool, default=False
            Whether this is the first layer
        omega_0 : float, default=30
            Frequency of the sine activation
        """
        super().__init__()
        self.omega_0 = omega_0
        self.is_first = is_first

        self.in_features = in_features
        self.linear = nn.Linear(in_features, out_features, bias=bias)

        with torch.no_grad():
            if self.is_first:
                self.linear.weight.uniform_(
                    -1 / self.in_features, 1 / self.in_features
                )
            else:
                limit = np.sqrt(6 / self.in_features) / self.omega_0
                self.linear.weight.uniform_(-limit, limit)

    def forward(self, x):
        return torch.sin(self.omega_0 * self.linear(x))


class SIREN(nn.Module):
    """
    SIREN (Sinusoidal Representation Networks) implementation.
    """

    def __init__(
        self,
        in_features: int = 3,
        out_features: int = 1,
        hidden_features: int = 256,
        hidden_layers: int = 3,
        outermost_linear: bool = True,
        omega_0: float = 30,
        omega_hidden: float = 30,
    ) -> None:
        """
        Initialize SIREN network with specified architecture.

        Parameters
        ----------
        in_features : int, default=3
            Number of input features
        out_features : int, default=1
            Number of output features
        hidden_features : int, default=256
            Width of hidden layers
        hidden_layers : int, default=3
            Number of hidden layers
        outermost_linear : bool, default=True
            Whether to use linear activation in final layer
        omega_0 : float, default=30
            Frequency for first sine layer
        omega_hidden : float, default=30
            Frequency for subsequent sine layers
        """
        super().__init__()

        self.net = []
        self.net.append(
            SineLayer(
                in_features, hidden_features, is_first=True, omega_0=omega_0
            )
        )

        for i in range(hidden_layers - 1):
            self.net.append(
                SineLayer(
                    hidden_features,
                    hidden_features,
                    is_first=False,
                    omega_0=omega_hidden,
                )
            )

        if outermost_linear:
            final_linear = nn.Linear(hidden_features, out_features)
            with torch.no_grad():
                limit = np.sqrt(6 / hidden_features) / omega_hidden
                final_linear.weight.uniform_(-limit, limit)
            self.net.append(final_linear)
        else:
            self.net.append(
                SineLayer(
                    hidden_features,
                    out_features,
                    is_first=False,
                    omega_0=omega_hidden,
                )
            )

        self.net = nn.Sequential(*self.net)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass through the SIREN network.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor of shape (..., in_features)

        Returns
        -------
        torch.Tensor
            Output tensor of shape (..., out_features)
        """
        return self.net(x)



from enum import Enum
from unittest.mock import MagicMock, patch

import pytest

from climatrix.reconstruct.idw import IDWReconstructor
from climatrix.reconstruct.type import ReconstructionType


class TestReconstructionType:

    def test_enum_idw_value(self):
        assert ReconstructionType.IDW.value == IDWReconstructor

    @pytest.mark.skipif(
        not hasattr(ReconstructionType, "OK"),
        reason="OrdinaryKrigingReconstructor is not available",
    )
    def test_enum_ok_value(self):
        from climatrix.reconstruct.kriging import OrdinaryKrigingReconstructor

        assert ReconstructionType.OK.value == OrdinaryKrigingReconstructor

    @pytest.mark.skipif(
        not hasattr(ReconstructionType, "SINET"),
        reason="SiNETReconstructor is not available",
    )
    def test_enum_sinet_value(self):
        from climatrix.reconstruct.sinet.sinet import SiNETReconstructor

        assert ReconstructionType.SINET.value == SiNETReconstructor

    @pytest.mark.skipif(
        not hasattr(ReconstructionType, "SIREN"),
        reason="SIRENReconstructor is not available",
    )
    def test_enum_siren_value(self):
        from climatrix.reconstruct.siren.siren import SIRENReconstructor

        assert ReconstructionType.SIREN.value == SIRENReconstructor

    def test_missing_method(self):
        with pytest.raises(
            ValueError,
            match="'invalid_method' is not a valid ReconstructionType",
        ):
            ReconstructionType("invalid_method")

    def test_get_method_with_enum_instance(self):
        rt = ReconstructionType.IDW
        result = ReconstructionType.get(rt)
        assert result is rt
        assert result == ReconstructionType.IDW

    def test_get_method_with_string_case_insensitive(self):
        result = ReconstructionType.get("IDW")
        assert result == ReconstructionType.IDW

        result = ReconstructionType.get("idw")
        assert result == ReconstructionType.IDW

        result = ReconstructionType.get("IdW")
        assert result == ReconstructionType.IDW

    def test_get_method_with_invalid_string(self):
        with pytest.raises(
            ValueError, match="Unknown reconstruction type: invalid_type"
        ):
            ReconstructionType.get("invalid_type")

    def test_get_method_with_invalid_type(self):
        with pytest.raises(TypeError, match="Invalid reconstruction type"):
            ReconstructionType.get(123)

        with pytest.raises(TypeError, match="Invalid reconstruction type"):
            ReconstructionType.get(None)

    def test_get_method_with_all_valid_keys(self):
        for rt in ReconstructionType:
            assert ReconstructionType.get(rt) == rt

            assert ReconstructionType.get(rt.name) == rt

    def test_enum_lifecycle(self):
        rt = ReconstructionType.get("idw")
        assert rt == ReconstructionType.IDW

        reconstructor_class = rt.value
        assert reconstructor_class == IDWReconstructor

        with patch.object(IDWReconstructor, "__init__", return_value=None):
            instance = reconstructor_class()
            assert isinstance(instance, IDWReconstructor)

    def test_list_method(self):
        available_types = ReconstructionType.list()
        assert isinstance(available_types, list)
        "IDW" in available_types



import xarray as xr

import climatrix as cm  # noqa # pylint: disable=unused-import

dset = xr.open_dataset(
    "/home/jakub/tul/research/climatrix/data/static-era5-land.nc"
)
europe = dset.cm.to_signed_longitude().subset(
    north=71, south=36, west=-24, east=35
)
sum = europe + 1000
sum.plot()



import glob

import numpy as np
import pandas as pd
from rich.console import Console
from rich.table import Table

ROOT_EXP_DIR = "results/ordinary_kriging"
UNIFORM_EXP_DIR = f"{ROOT_EXP_DIR}/uniform"
NORMAL_EXP_DIR = f"{ROOT_EXP_DIR}/normal"

normal_metric_files = glob.glob(f"{NORMAL_EXP_DIR}/*/metrics.csv")


def analyse_uniform():
    uniform_metric_files = glob.glob(f"{UNIFORM_EXP_DIR}/*/metrics.csv")
    if len(uniform_metric_files) == 0:
        return
    uniform_metrics = pd.concat(
        [pd.read_csv(f) for f in uniform_metric_files], ignore_index=True
    )
    table = Table(
        title="Metrics for reconstruction from uniformly sampled "
        "data (Ordinary Kriging)"
    )
    table.add_column("Metric")
    table.add_column("Mean value")
    table.add_column("Std value")
    table.add_column("Std error")

    for col in uniform_metrics.columns:
        std_err = uniform_metrics[col].std() / np.sqrt(len(uniform_metrics))
        table.add_row(
            col,
            f"{uniform_metrics[col].mean():.4f}",
            f"{uniform_metrics[col].std():.4f}",
            f"{std_err:.4f}",
        )
    console = Console()
    console.print(table)


def analyse_normal():
    normal_metric_files = glob.glob(f"{NORMAL_EXP_DIR}/*/metrics.csv")
    if len(normal_metric_files) == 0:
        return
    normal_metrics = pd.concat(
        [pd.read_csv(f) for f in normal_metric_files], ignore_index=True
    )
    table = Table(
        title="Metrics for reconstruction from uniformly sampled "
        "data (Ordinary Kriging)"
    )
    for col in normal_metrics.columns:
        table.add_column(f"mean {col}")
        table.add_column(f"std {col}")

    for col in normal_metrics.columns:
        table.add_row(
            f"{normal_metrics[col].mean():.4f}",
            f"{normal_metrics[col].std():.4f}",
        )
    console = Console()
    console.print(table)


if __name__ == "__main__":
    analyse_uniform()
    analyse_normal()
